#
# Copyright 2012 Michal Čihař
# Copyright 2014 Luca De Petrillo
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Module for handling Android String and Plurals resource files."""

from __future__ import annotations

import copy
import os
import re
from xml.parsers.expat import XML_PARAM_ENTITY_PARSING_NEVER, ParserCreate

from lxml import etree

from translate.lang import data
from translate.misc.multistring import multistring
from translate.storage import base, lisa

WHITESPACE = {" ", "\n", "\t"}  # Whitespace that we collapse.
ESCAPE_NEWLINE = {"n", "N"}
ESCAPE_TAB = {"t", "T"}
ESCAPE_PLAIN = {" ", '"', "'", "@", "?", "\\"}
MULTIWHITESPACE = re.compile(r"[ \n\t]{2}(?!\\n)")
QUOTED_STRING = re.compile(r'((?<!\\)"(?:\\"|[^"])*(?<!\\)")')
UNICODE_ESCAPE = re.compile(r"\\u([a-fA-F0-9]{4})")
CHAR_ESCAPE = re.compile(r"\\(.)")
WHITESPACE_RE = re.compile(r"\s+")

ESCAPE_TRANSLATE = str.maketrans(
    {
        "\\": "\\\\",
        "\n": "\\n",
        "\t": "\\t",
        "'": "\\'",
        '"': '\\"',
    }
)


class DecodingXMLParser:
    """
    XML parser that processes non CDATA strings.

    - all XML markup from given depth is output as is
    - CDATA elements are output as is
    - process_string() is called on all other textual content
    """

    EMIT_DEPTH = 1
    FOREGIN_DTD = True

    def __init__(self, text: str):
        self.text = text.encode("utf-8")
        self.output = []
        self.emit_start = None
        self.decoded_emit = None
        self.character_data = False
        self.raw_string = True
        self.in_string = False
        self.do_cleanup = False
        self.depth = 0
        self.parser = parser = ParserCreate()
        if self.FOREGIN_DTD:
            parser.UseForeignDTD(self.FOREGIN_DTD)
        parser.SetParamEntityParsing(XML_PARAM_ENTITY_PARSING_NEVER)
        parser.StartElementHandler = self.StartElementHandler
        parser.EndElementHandler = self.EndElementHandler
        parser.CharacterDataHandler = self.CharacterDataHandler
        parser.StartCdataSectionHandler = self.StartCdataSectionHandler
        parser.EndCdataSectionHandler = self.EndCdataSectionHandler
        parser.DefaultHandler = self.DefaultHandler

    @staticmethod
    def decode_unicode_escapes(match: re.Match) -> str:
        escaped = match.group(1)
        # Convert from hex to character
        return chr(int(escaped, 16))

    @staticmethod
    def decode_escapes(match: re.Match) -> str:
        escaped = match.group(1)
        if escaped in ESCAPE_NEWLINE:
            return "\n"  # an actual newline
        if escaped in ESCAPE_TAB:
            return "\t"  # an actual tab
        if escaped in ESCAPE_PLAIN:
            # Plain string to escape
            return escaped
        # All others, remove, like Android does as well.
        return ""

    @staticmethod
    def strip_part(text: str) -> str:
        if text.startswith('"'):
            # Missing closing quote is gracefully ignored
            return text.removeprefix('"').removesuffix('"')
        # Consolidate whitespace
        return WHITESPACE_RE.sub(" ", text)

    @classmethod
    def process_string(cls, content: str) -> tuple[str, bool, bool]:
        # Skip processing for a blank string
        if not content:
            return "", False, False

        # Extract quoted parts, remove blank ones
        parts = [part for part in QUOTED_STRING.split(content) if part]

        # Detect possible cleanups
        cleanup_start = not parts[0].startswith('"')
        cleanup_end = not parts[-1].startswith('"')

        # Collapse whitespace
        stripped_parts = [cls.strip_part(part) for part in parts]
        # Merge this back to a single string
        text = "".join(stripped_parts)

        # Handle unicode escapes
        unescaped_text = UNICODE_ESCAPE.sub(cls.decode_unicode_escapes, text)
        # Handle other escapes
        unescaped_text = CHAR_ESCAPE.sub(cls.decode_escapes, unescaped_text)

        # Detect escaping at the start and end of a string (we actually care only for
        # whitespace escapes, but detecting more does not matter)
        if text and unescaped_text and text[0] != unescaped_text[0]:
            cleanup_start = False
        if text and unescaped_text and text[-1] != unescaped_text[-1]:
            cleanup_end = False

        return unescaped_text, cleanup_start, cleanup_end

    def cleanup_text(self, text: str) -> str:
        """Remove leading whitespace from text."""
        if not self.output:
            return text.lstrip()
        return text

    def emit(self):
        if self.emit_start is not None:
            text = self.text[self.emit_start : self.parser.CurrentByteIndex].decode(
                "utf-8"
            )
            self.do_cleanup = not self.raw_string
            if not self.raw_string:
                text, cleanup_start, self.do_cleanup = self.process_string(text)
                if cleanup_start:
                    text = self.cleanup_text(text)
            self.output.append(text)
            self.emit_start = None
        self.character_data = False
        self.raw_string = True
        self.in_string = False

    def StartElementHandler(self, _name, _attributes):
        self.emit()
        if self.depth >= self.EMIT_DEPTH:
            self.emit_start = self.parser.CurrentByteIndex
        self.depth += 1

    def EndElementHandler(self, _name):
        self.emit()
        if self.depth >= self.EMIT_DEPTH:
            self.emit_start = self.parser.CurrentByteIndex
        self.depth -= 1

    def CharacterDataHandler(self, _data):
        if not self.character_data and not self.in_string:
            self.emit()
            self.emit_start = self.parser.CurrentByteIndex
            self.raw_string = False
            self.in_string = True

    def StartCdataSectionHandler(self):
        self.emit()
        self.character_data = True
        self.emit_start = self.parser.CurrentByteIndex

    def EndCdataSectionHandler(self):
        self.character_data = False

    def DefaultHandler(self, _data):
        if self.depth >= self.EMIT_DEPTH:
            self.emit()
            self.emit_start = self.parser.CurrentByteIndex

    def parse(self) -> str:
        self.parser.Parse(self.text, True)
        if self.depth >= self.EMIT_DEPTH:
            self.emit()
        # Remove trailing whitespace from text
        if self.do_cleanup and self.output:
            self.output[-1] = self.output[-1].rstrip()
        return "".join(self.output)


class EncodingXMLParser(DecodingXMLParser):
    EMIT_DEPTH = 0
    FOREGIN_DTD = False

    @staticmethod
    def process_string(text: str) -> tuple[str, bool, bool]:
        return (
            AndroidResourceUnit.escape(text, quote_wrapping_whitespaces=False),
            False,
            False,
        )

    def parse(self) -> str:
        self.emit_start = 0
        return super().parse()


class AndroidResourceUnit(base.TranslationUnit):
    """A single entry in the Android String resource file."""

    SINGULAR_TAG = "string"
    PLURAL_TAG = "plurals"

    @classmethod
    def createfromxmlElement(cls, element):
        term = None
        # Actually this class supports only plurals and string tags
        if element.tag in {cls.PLURAL_TAG, cls.SINGULAR_TAG}:
            term = cls(None, xmlelement=element)
        return term

    def __init__(self, source, empty=False, xmlelement=None, **kwargs):
        if xmlelement is not None:
            self.xmlelement = xmlelement
        elif self.hasplurals(source):
            self.xmlelement = etree.Element(self.PLURAL_TAG)
        else:
            self.xmlelement = etree.Element(self.SINGULAR_TAG)
        if source is not None:
            self.setid(source)
        super().__init__(source)

    def istranslatable(self):
        return bool(self.getid()) and self.xmlelement.get("translatable") != "false"

    def marktranslatable(self, value: bool) -> None:
        return self.xmlelement.set("translatable", "true" if value else "false")

    def isblank(self):
        return not bool(self.getid())

    def getid(self):
        return self.xmlelement.get("name")

    def setid(self, newid):
        return self.xmlelement.set("name", newid)

    def getcontext(self):
        return self.xmlelement.get("name")

    @staticmethod
    def xml_escape_space(matchobj):
        return matchobj.group(0).replace("  ", r" \u0020")

    @classmethod
    def escape(
        cls, text: str | None, quote_wrapping_whitespaces: bool = True
    ) -> str | None:
        """
        Escape all the characters which need to be escaped in an Android XML
        file.

        :param text: Text to escape
        :param quote_wrapping_whitespaces: If True, heading and trailing
               whitespaces will be quoted placing the entire resulting text in
               double quotes.
        """
        if text is None:
            return None
        if len(text) == 0:
            return ""

        # Escape XML chars and whitespace
        text = text.translate(ESCAPE_TRANSLATE)

        # @ and ? needs to be escaped at start as this would be interpreted
        # as string/style references
        if text.startswith(("@", "?")):
            text = f"\\{text}"

        # Quote strings with more whitespace
        multispace = MULTIWHITESPACE.search(text)
        if quote_wrapping_whitespaces and (
            text[0] in WHITESPACE or text[-1] in WHITESPACE or multispace
        ):
            return f'"{text}"'
        # In xml multispace
        if not quote_wrapping_whitespaces and multispace:
            return MULTIWHITESPACE.sub(cls.xml_escape_space, text)

        return text

    @base.TranslationUnit.source.getter
    def source(self):
        if super().source is None:
            return self.target
        return super().source

    def get_xml_text_value(self, xmltarget):
        raw_xml = etree.tostring(xmltarget, encoding="unicode", with_tail=False)
        # Unescape as plain text in case there is no XML markup. Ufortunately
        # CDATA elements are not listed as childs in lxml, so test for it in raw XML.
        if len(xmltarget) == 0 and "<![CDATA[" not in raw_xml:
            if xmltarget.text is None:
                return ""
            text, cleanup_start, cleanup_end = DecodingXMLParser.process_string(
                xmltarget.text
            )
            if cleanup_start:
                text = text.lstrip()
            if cleanup_end:
                text = text.rstrip()
            return text

        # Use decoding parser to keep XML entitities, CDATA while processing Android escaping
        parser = DecodingXMLParser(raw_xml)
        return parser.parse()

    def set_xml_text_plain(self, target, xmltarget):
        # Remove possible old elements
        for x in xmltarget.iterchildren():
            xmltarget.remove(x)
        # Handle text only
        xmltarget.text = self.escape(target)

    def set_xml_text_value(self, target, xmltarget):
        if "<" in target or "&" in target:
            # Try to handle it as legacy XML
            parser = etree.XMLParser(strip_cdata=False, resolve_entities=False)
            if self._store is not None:
                cloned_doc = copy.deepcopy(self._store.document)
                cloned_root = cloned_doc.getroot()
                cloned_root.clear()
                # Add content to the element so that we get a real closing tag
                cloned_root.text = " "
                template = etree.tostring(
                    cloned_doc,
                    encoding="unicode",
                    doctype=copy.copy(self._store.XMLdoctype),
                )
            else:
                template = "<resources></resources>"
            newstring = template.replace(
                "</resources>",
                f"<{xmltarget.tag}>{target}</{xmltarget.tag}></resources>",
            )
            try:
                etree.fromstring(newstring, parser)
            except Exception:
                self.set_xml_text_plain(target, xmltarget)
            else:
                # Escape text parts
                encoding_parser = EncodingXMLParser(newstring)
                newstring = encoding_parser.parse()
                newtree = etree.fromstring(newstring, parser)[0]
                # Copy existing attributes
                for attribute, value in xmltarget.items():
                    newtree.attrib[attribute] = value
                # Update text
                parent = xmltarget.getparent()
                # Parent can be none when operating on Unit only without a storage
                if parent is not None:
                    parent.replace(xmltarget, newtree)
                # Update unit xmlelement if needed (plurals are inner elements here)
                if xmltarget == self.xmlelement:
                    self.xmlelement = newtree
        else:
            self.set_xml_text_plain(target, xmltarget)

    def gettargetlanguage(self):
        if self._store is None:
            return "en"
        return super().gettargetlanguage()

    def get_plural_tags(self):
        locale = self.gettargetlanguage()
        if not locale:
            return data.plural_tags["en"]
        # Handle b+ style language codes
        locale = locale.removeprefix("b+")
        locale = locale.replace("_", "-").replace("+", "-").split("-")[0]
        return data.plural_tags.get(locale, data.plural_tags["en"])

    @property
    def target(self):
        if self.xmlelement.tag != self.PLURAL_TAG:
            return self.get_xml_text_value(self.xmlelement)
        plurals = {
            entry.get("quantity"): self.get_xml_text_value(entry)
            for entry in self.xmlelement.iterchildren("item")
        }
        plural_tags = self.get_plural_tags()
        return multistring([plurals.get(tag, "") for tag in plural_tags])

    @target.setter
    def target(self, target):
        if self.hasplurals(self.source) or self.hasplurals(target):
            # Fix the root tag if mismatching
            if self.xmlelement.tag != self.PLURAL_TAG:
                old_id = self.getid()
                self.xmlelement = etree.Element(self.PLURAL_TAG)
                self.setid(old_id)

            plural_tags = self.get_plural_tags()

            # Sync plural_strings elements to plural_tags count.
            plural_strings = self.sync_plural_count(target, plural_tags)

            # Rebuild plurals.
            for entry in self.xmlelement.iterchildren():
                self.xmlelement.remove(entry)

            self.xmlelement.text = "\n    "

            # Include "other" as copy of "many" if "other" is not present. This avoids crashes
            # of Android builts with broken plurals handling.
            if "other" not in plural_tags and "many" in plural_tags:
                # Create copy here to avoid modifications to laguage.data
                plural_tags = [*plural_tags, "other"]
                plural_strings.append(plural_strings[-1])

            for plural_tag, plural_string in zip(plural_tags, plural_strings):
                item = etree.Element("item")
                item.set("quantity", plural_tag)
                self.xmlelement.append(item)
                self.set_xml_text_value(plural_string, item)
        else:
            # Fix the root tag if mismatching
            if self.xmlelement.tag != self.SINGULAR_TAG:
                old_id = self.getid()
                self.xmlelement = etree.Element(self.SINGULAR_TAG)
                self.setid(old_id)

            self.set_xml_text_value(target, self.xmlelement)

        self._rich_target = None
        self._target = target

    def getlanguageNode(self, lang=None, index=None):
        return self.xmlelement

    # Notes are handled as previous sibling comments.
    def addnote(self, text, origin=None, position="append"):
        if origin in {"programmer", "developer", "source code", None}:
            self.xmlelement.addprevious(etree.Comment(text))
        else:
            super().addnote(text, origin=origin, position=position)

    def getnotes(self, origin=None):
        if origin in {"programmer", "developer", "source code", None}:
            comments = []
            if self.xmlelement is not None:
                prevSibling = self.xmlelement.getprevious()
                while (prevSibling is not None) and (prevSibling.tag is etree.Comment):
                    comments.insert(0, prevSibling.text)
                    prevSibling = prevSibling.getprevious()

            return "\n".join(comments)
        return super().getnotes(origin)

    def removenotes(self, origin=None):
        if (self.xmlelement is not None) and (self.xmlelement.getparent is not None):
            prevSibling = self.xmlelement.getprevious()
            while (prevSibling is not None) and (prevSibling.tag is etree.Comment):
                prevSibling.getparent().remove(prevSibling)
                prevSibling = self.xmlelement.getprevious()

        super().removenotes()

    def __str__(self):
        return etree.tostring(self.xmlelement, pretty_print=True, encoding="unicode")

    def __eq__(self, other):
        return str(self) == str(other)

    @staticmethod
    def hasplurals(thing):
        return isinstance(thing, (multistring, list))


class AndroidResourceFile(lisa.LISAfile):
    """Class representing an Android String resource file store."""

    UnitClass = AndroidResourceUnit
    Name = "Android String Resource"
    Mimetypes = ["application/xml"]
    Extensions = ["xml"]
    rootNode = "resources"
    bodyNode = "resources"
    XMLskeleton = """<?xml version="1.0" encoding="utf-8"?>
<resources></resources>"""
    XMLindent = {"indent": "    ", "leaves": {"string", "item"}, "trailing_eol": True}
    XMLuppercaseEncoding = False

    def initbody(self):
        """
        Initialises self.body so it never needs to be retrieved from the XML
        again.
        """
        self.namespace = self.document.getroot().nsmap.get(None, None)
        self.body = self.document.getroot()

    def parse(self, xml):
        """Populates this object from the given xml string."""
        if not hasattr(self, "filename"):
            self.filename = getattr(xml, "name", "")
        if hasattr(xml, "read"):
            xml.seek(0)
            posrc = xml.read()
            xml = posrc
        parser = etree.XMLParser(strip_cdata=False, resolve_entities=False)
        self.document = etree.fromstring(xml, parser).getroottree()
        self._encoding = self.document.docinfo.encoding
        self.initbody()
        assert self.document.getroot().tag == self.namespaced(self.rootNode)

        for entry in self.document.getroot().iterchildren():
            term = self.UnitClass.createfromxmlElement(entry)
            if term is not None:
                self.addunit(term, new=False)

    def gettargetlanguage(self):
        target_lang = super().gettargetlanguage()

        # If targetlanguage isn't set, we try to extract it from the filename path (if any).
        if target_lang is None and hasattr(self, "filename") and self.filename:
            # Android standards expect resource files to be in a directory named "values[-<lang>[-r<region>]]".
            parent_dir = os.path.split(os.path.dirname(self.filename))[1]
            match = re.search(r"^values-(\w*)", parent_dir)
            if match is not None:
                target_lang = match.group(1)
            elif parent_dir == "values":
                # If the resource file is inside the "values" directory, then it is the default/source language.
                target_lang = self.sourcelanguage

            # Cache it
            self.settargetlanguage(target_lang)

        return target_lang

    def addunit(self, unit, new=True):
        """
        Adds unit to the document.

        In addition to the standard addunit, it also tries to move
        namespace definitions to the top <resources> element.
        """
        newns = {}
        do_cleanup = False
        if new:
            # Include any possible new namespaces
            newns = self.body.nsmap
            for ns in unit.xmlelement.nsmap:
                if ns not in newns:
                    do_cleanup = True
                    newns[ns] = unit.xmlelement.nsmap[ns]

            # Detect if unit is using XML entities
            hasentity = False
            for xmlelement in unit.xmlelement.iterdescendants():
                if xmlelement.tag is etree.Entity:
                    hasentity = True
                    break

            # Copy doctype to include possibly new entities
            if hasentity:
                cloned_doc = copy.deepcopy(unit._store.document)
                cloned_doc.getroot().clear()
                self.XMLdoctype = etree.tostring(
                    cloned_doc, xml_declaration=False, encoding="unicode"
                ).rsplit("\n", 1)[0]

        super().addunit(unit, new)
        # Move aliased namespaces to the <resources> tag
        # The top_nsmap was introduced in LXML 3.5.0
        if do_cleanup:
            etree.cleanup_namespaces(self.body, top_nsmap=newns)

    def removeunit(self, unit):
        unit.removenotes()
        super().removeunit(unit)


class MOKOResourceUnit(AndroidResourceUnit):
    PLURAL_TAG = "plural"


class MOKOResourceFile(AndroidResourceFile):
    UnitClass = MOKOResourceUnit



#
# Copyright 2007-2009,2011 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
This module represents the Greek language.

.. seealso:: :wp:`Greek_language`
"""

import re

from translate.lang import common


class el(common.Common):
    """This class represents Greek."""

    # Greek uses ; as question mark and the middot instead
    sentenceend = ".!;…"

    sentencere = re.compile(
        rf"""
        (?s)        # make . also match newlines
        .*?         # anything, but match non-greedy
        [{sentenceend}]        # the puntuation for sentence ending
        \s+         # the spacing after the puntuation
        (?=[^a-zά-ώ\d])  # lookahead that next part starts with caps
        """,
        re.VERBOSE | re.UNICODE,
    )

    puncdict = {
        ";": "·",
        "?": ";",
    }

    # Valid latin characters for use as accelerators
    valid_latin_accel = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890"

    # Valid greek characters for use as accelerators (accented characters
    # and "ς" omitted)
    valid_greek_accel = "αβγδεζηθικλμνξοπρστυφχψωΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩ"

    # Valid accelerators
    validaccel = f"{valid_latin_accel}{valid_greek_accel}"



#
# Copyright 2008-2011 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Module for handling Qt linguist (.ts) files.

This will eventually replace the older ts.py which only supports the older
format. While converters haven't been updated to use this module, we retain
both.

`TS file format 4.3 <http://doc.qt.io/archives/4.3/linguist-ts-file-format.html>`_,
`4.8 <http://doc.qt.io/qt-4.8/linguist-ts-file-format.html>`_,
`5 <http://doc.qt.io/qt-5/linguist-ts-file-format.html>`_.
`Example <http://svn.ez.no/svn/ezcomponents/trunk/Translation/docs/linguist-format.txt>`_.

`Specification of the valid variable entries <http://doc.qt.io/qt-5/qstring.html#arg>`_,
`2 <http://doc.qt.io/qt-5/qstring.html#arg-2>`_
"""

from lxml import etree

from translate.lang import data
from translate.misc.multistring import multistring
from translate.misc.xml_helpers import safely_set_text
from translate.storage import lisa
from translate.storage.placeables import general
from translate.storage.workflow import StateEnum as state

# TODO: handle translation types


# Encode some characters same as lupdate, this matches tsProtect implementation:
#
# - quote and apostrophe
# - any whitespace character above 0x7f
#
# The latter can be generated using:
# >>> for i in  range(sys.maxunicode):
# ...     if i > 0x7f and chr(i).isspace():
# ...         print(f'   {chr(i)!r}: "&#x{i:x};",')

OUTPUT_TRANS = str.maketrans(
    {
        "'": "&apos;",
        '"': "&quot;",
        "\xa0": "&#xa0;",
        "\u1680": "&#x1680;",
        "\u2000": "&#x2000;",
        "\u2001": "&#x2001;",
        "\u2002": "&#x2002;",
        "\u2003": "&#x2003;",
        "\u2004": "&#x2004;",
        "\u2005": "&#x2005;",
        "\u2006": "&#x2006;",
        "\u2007": "&#x2007;",
        "\u2008": "&#x2008;",
        "\u2009": "&#x2009;",
        "\u200a": "&#x200a;",
        "\u2028": "&#x2028;",
        "\u2029": "&#x2029;",
        "\u202f": "&#x202f;",
        "\u205f": "&#x205f;",
        "\u3000": "&#x3000;",
    }
)


class tsunit(lisa.LISAunit):
    """A single term in the TS file."""

    rootNode = "message"
    languageNode = "source"
    textNode = ""
    namespace = ""
    rich_parsers = general.parsers

    S_OBSOLETE = state.OBSOLETE
    S_UNTRANSLATED = state.EMPTY
    S_FUZZY = state.NEEDS_WORK
    S_TRANSLATED = state.UNREVIEWED

    statemap = {
        "obsolete": S_OBSOLETE,
        "unfinished": S_FUZZY,
        "": S_TRANSLATED,
        None: S_TRANSLATED,
    }
    """This maps the unit "type" attribute to state."""

    STATE = {
        S_OBSOLETE: (state.OBSOLETE, state.EMPTY),
        S_UNTRANSLATED: (state.EMPTY, state.NEEDS_WORK),
        S_FUZZY: (state.NEEDS_WORK, state.UNREVIEWED),
        S_TRANSLATED: (state.UNREVIEWED, state.MAX),
    }

    statemap_r = {i[1]: i[0] for i in statemap.items()}
    _context = None
    _locations = None

    def createlanguageNode(self, lang, text, purpose):
        """Returns an xml Element setup with given parameters."""
        assert purpose
        if purpose == "target":
            purpose = "translation"
        langset = etree.Element(self.namespaced(purpose))
        # TODO: check language
        # lisa.setXMLlang(langset, lang)

        safely_set_text(langset, text)
        return langset

    def _getsourcenode(self):
        return self.xmlelement.find(self.namespaced(self.languageNode))

    def _gettargetnode(self):
        result = self.xmlelement.find(self.namespaced("translation"))
        if result is not None:
            return result
        return etree.SubElement(self.xmlelement, self.namespaced("translation"))

    def getlanguageNodes(self):
        """We override this to get source and target nodes."""
        return [
            n for n in [self._getsourcenode(), self._gettargetnode()] if n is not None
        ]

    @lisa.LISAunit.source.getter
    def source(self):
        sourcenode = self._getsourcenode()
        if sourcenode is None:
            return None
        # TODO: support <byte>. See bug 528.
        text = sourcenode.text
        if self.hasplural():
            return multistring([text])
        return text

    @property
    def target(self):
        targetnode = self._gettargetnode()
        if self.hasplural():
            numerus_nodes = targetnode.findall(self.namespaced("numerusform"))
            return multistring([node.text or "" for node in numerus_nodes])
        return targetnode.text or ""

    @target.setter
    def target(self, target):
        # This is a fairly destructive implementation. Don't assume that this
        # is necessarily correct in all regards, but it does deal with a lot of
        # cases. It is hard to deal with plurals.
        #
        # Firstly deal with reinitialising to None or setting to identical
        # string.
        self._rich_target = None
        if self.target == target:
            return
        strings = []
        if isinstance(target, multistring):
            strings = target.strings
        elif isinstance(target, list):
            strings = target
        else:
            strings = [target]
        targetnode = self._gettargetnode()
        type = targetnode.get("type")
        targetnode.clear()
        if type:
            targetnode.set("type", type)
        if self.hasplural() or len(strings) > 1:
            self.xmlelement.set("numerus", "yes")
            for string in strings:
                numerus = etree.SubElement(targetnode, self.namespaced("numerusform"))
                safely_set_text(numerus, string or "")
        else:
            safely_set_text(targetnode, target or "")

    def hasplural(self):
        return self.xmlelement.get("numerus") == "yes"

    def addnote(self, text, origin=None, position="append"):
        """Add a note specifically in the appropriate *comment* tag."""
        current_notes = self.getnotes(origin)
        self.removenotes(origin)
        if origin in {"programmer", "developer", "source code"}:
            note = etree.SubElement(self.xmlelement, self.namespaced("extracomment"))
        else:
            note = etree.SubElement(
                self.xmlelement, self.namespaced("translatorcomment")
            )
        if position == "append":
            safely_set_text(
                note, "\n".join(item for item in [current_notes, text.strip()] if item)
            )
        else:
            safely_set_text(note, text.strip())

    def getnotes(self, origin=None):
        # TODO: consider only responding when origin has certain values
        comments = []
        if origin in {"programmer", "developer", "source code", None}:
            notenode = self.xmlelement.find(self.namespaced("extracomment"))
            if notenode is not None and notenode.text is not None:
                comments.append(notenode.text)
        if origin in {"translator", None}:
            notenode = self.xmlelement.find(self.namespaced("translatorcomment"))
            if notenode is not None and notenode.text is not None:
                comments.append(notenode.text)
        return "\n".join(comments)

    def removenotes(self, origin=None):
        """Remove all the translator notes."""
        if origin in {"programmer", "developer", "source code", None}:
            note = self.xmlelement.find(self.namespaced("extracomment"))
            if note is not None:
                self.xmlelement.remove(note)
        if origin in {"translator", None}:
            note = self.xmlelement.find(self.namespaced("translatorcomment"))
            if note is not None:
                self.xmlelement.remove(note)

    def _gettype(self):
        """Returns the type of this translation."""
        return self._gettargetnode().get("type")

    def _settype(self, value=None):
        """Set the type of this translation."""
        if value:
            self._gettargetnode().set("type", value)
        elif self._gettype():
            # lxml recommends against using .attrib, but there seems to be no
            # other way
            self._gettargetnode().attrib.pop("type")

    def isreview(self):
        """States whether this unit needs to be reviewed."""
        return self._gettype() == "unfinished"

    def isfuzzy(self):
        return self._gettype() == "unfinished" and bool(self.target)

    def markfuzzy(self, value=True):
        if value:
            self._settype("unfinished")
        else:
            self._settype(None)

    def getid(self):
        context_name = self.getcontext()
        if self.source is None and context_name is None:
            return None

        # XXX: context_name is not supposed to be able to be None (the <name>
        # tag is compulsary in the <context> tag)
        if context_name is not None:
            if self.source:
                return context_name + self.source
            return context_name
        return self.source

    def istranslatable(self):
        # Found a file in the wild with no context and an empty source. This
        # served as a header, so let's classify this as not translatable.
        # http://bibletime.svn.sourceforge.net/viewvc/bibletime/trunk/bibletime/i18n/messages/bibletime_ui.ts
        # Furthermore, let's decide to handle obsolete units as untranslatable
        # like we do with PO.
        return bool(self.getid()) and not self.isobsolete()

    def getcontextname(self):
        parent = self.xmlelement.getparent()
        if parent is None:
            if self._context:
                return self._context
            return None
        context = parent.find("name")
        if context is None:
            return None
        return context.text

    def setcontext(self, value):
        if value == self.getcontextname():
            return
        parent = self.xmlelement.getparent()
        if parent is None:
            self._context = value
        # TODO: if unit is inserted to ts store
        # it should put unit in correct context
        # if unit is already in store it should move
        # and adjust store as necessary

    def getcontext(self):
        contexts = [self.getcontextname()]
        commentnode = self.xmlelement.find(self.namespaced("comment"))
        if commentnode is not None and commentnode.text is not None:
            contexts.append(commentnode.text)
        message_id = self.xmlelement.get("id")
        if message_id is not None:
            contexts.append(message_id)
        contexts = filter(None, contexts)
        return "\n".join(contexts)

    def addlocation(self, location):
        self._locations = None
        newlocation = etree.SubElement(self.xmlelement, self.namespaced("location"))
        try:
            filename, line = location.split(":", 1)
        except ValueError:
            filename = location
            line = None
        newlocation.set("filename", filename)
        if line is not None:
            newlocation.set("line", line)

    def parse_locations(self):
        location_tags = self.xmlelement.iterfind(self.namespaced("location"))
        locations = []
        last_location = None
        for location_tag in location_tags:
            location = location_tag.get("filename")
            line = location_tag.get("line")
            if line is not None and line.startswith(("+", "-")):
                # Relative locations
                if last_location is None and (
                    previous_unit := self.get_previous_unit()
                ):
                    last_location = previous_unit.get_last_location()

                offset = 0
                if last_location:
                    if not location:
                        location = last_location[0]
                    offset = last_location[1]
                line = offset + int(line)
            if location or line:
                last_location = (location, line)
                locations.append(last_location)
        self._locations = locations

    def get_last_location(self):
        if self._locations is None:
            self.parse_locations()

        if not self._locations:
            previous = self.get_previous_unit()
            if previous is None:
                return None
            return previous.get_last_location()

        return self._locations[-1]

    def get_previous_unit(self):
        found = None
        for pos, unit in enumerate(self._store.units):
            # Use is here to compare objects as __eq__ implementation in
            # LISAUnit might give unexpected results
            if unit is self:
                found = pos
                break
        if not found or found == 0:
            return None
        return self._store.units[found - 1]

    def getlocations(self):
        if self._locations is None:
            self.parse_locations()

        return [
            f"{location}{':' if location else ''}{line}"
            for location, line in self._locations
        ]

    def merge(self, otherunit, overwrite=False, comments=True, authoritative=False):
        super().merge(otherunit, overwrite, comments)
        # TODO: check if this is necessary:
        if otherunit.isfuzzy():
            self.markfuzzy()
        else:
            self.markfuzzy(False)

    def isobsolete(self):
        return self._gettype() in {"obsolete", "vanished"}

    def get_state_n(self):
        type = self._gettype()
        if type == "unfinished":
            # We want to distinguish between fuzzy and untranslated, which the
            # format doesn't really do
            if self.target:
                return self.S_FUZZY
            return self.S_UNTRANSLATED
        if type == "vanished":
            return self.S_OBSOLETE
        return self.statemap[type]

    def set_state_n(self, value):
        if value not in self.statemap_r:
            value = self.get_state_id(value)

        if value == self.S_UNTRANSLATED:
            # No real way of representing that in the format, so we just
            # handle it the same as unfinished
            value = self.S_FUZZY
        self._settype(self.statemap_r[value])


class tsfile(lisa.LISAfile):
    """Class representing a TS file store."""

    UnitClass = tsunit
    Name = "Qt Linguist Translation File"
    Mimetypes = ["application/x-linguist"]
    Extensions = ["ts"]
    rootNode = "TS"
    # We will switch out .body to fit with the context we are working on
    bodyNode = "context"
    XMLskeleton = """<!DOCTYPE TS>
<TS>
</TS>
"""
    XMLindent = {"indent": "    ", "skip": {"TS"}, "toplevel": False}
    # For conformance with Qt output, write XML declaration with double quotes
    XMLuppercaseEncoding = False
    namespace = ""

    def __init__(self, *args, **kwargs):
        self._contextname = None
        self.last_location = None
        super().__init__(*args, **kwargs)

    def initbody(self):
        """Initialises self.body."""
        self.namespace = self.document.getroot().nsmap.get(None, None)
        self.header = self.document.getroot()
        if self._contextname:
            self.body = self._getcontextnode(self._contextname)
        else:
            self.body = self.document.getroot()

    def getsourcelanguage(self):
        """
        Get the source language for this .ts file.

        The 'sourcelanguage' attribute was only added to the TS format in
        Qt v4.5. We return 'en' if there is no sourcelanguage set.

        We don't implement setsourcelanguage as users really shouldn't be
        altering the source language in .ts files, it should be set correctly
        by the extraction tools.

        :return: ISO code e.g. af, fr, pt_BR
        :rtype: String
        """
        lang = data.normalize_code(self.header.get("sourcelanguage", "en"))
        if lang == "en-us":
            return "en"
        return lang

    def gettargetlanguage(self):
        """
        Get the target language for this .ts file.

        :return: ISO code e.g. af, fr, pt_BR
        :rtype: String
        """
        return data.normalize_code(self.header.get("language"))

    def settargetlanguage(self, targetlanguage):
        """
        Set the target language for this .ts file to *targetlanguage*.

        :param targetlanguage: ISO code e.g. af, fr, pt_BR
        :type targetlanguage: String
        """
        if targetlanguage:
            self.header.set("language", targetlanguage)

    def _createcontext(self, contextname, comment=None):
        """Creates a context node with an optional comment."""
        context = etree.SubElement(
            self.document.getroot(), self.namespaced(self.bodyNode)
        )
        name = etree.SubElement(context, self.namespaced("name"))
        safely_set_text(name, contextname)
        if comment:
            comment_node = etree.SubElement(context, "comment")
            safely_set_text(comment_node, comment)
        return context

    def _getcontextname(self, contextnode):
        """Returns the name of the given context node."""
        return contextnode.find(self.namespaced("name")).text

    def _getcontextnames(self):
        """Returns all contextnames in this TS file."""
        contextnodes = self.document.findall(self.namespaced("context"))
        return [self.getcontextname(contextnode) for contextnode in contextnodes]

    def _getcontextnode(self, contextname):
        """Returns the context node with the given name."""
        contextnodes = self.document.findall(self.namespaced("context"))
        for contextnode in contextnodes:
            if self._getcontextname(contextnode) == contextname:
                return contextnode
        return None

    def addunit(
        self, unit, new=True, contextname=None, comment=None, createifmissing=True
    ):
        """
        Adds the given unit to the last used body node (current context).

        If the contextname is specified, switch to that context (creating it if
        allowed by createifmissing).
        """
        if contextname is None:
            contextname = unit.getcontextname()

        if self._contextname != contextname and not self._switchcontext(
            contextname, comment, createifmissing
        ):
            return None
        super().addunit(unit, new)
        #        lisa.setXMLspace(unit.xmlelement, "preserve")
        return unit

    def _switchcontext(self, contextname, comment, createifmissing=False):
        """
        Switch the current context to the one named contextname, optionally
        creating it if it doesn't exist.
        """
        self._contextname = contextname
        contextnode = self._getcontextnode(contextname)
        if contextnode is None:
            if not createifmissing:
                return False
            contextnode = self._createcontext(contextname, comment)

        self.body = contextnode
        return self.body is not None

    def nplural(self):
        code = self.header.get("language").lower().replace("-", "_").split("_")[0]
        if code in data.qt_plural_tags:
            return len(data.qt_plural_tags[code])
        lang = data.get_language(self.header.get("language"))
        if lang is None:
            return 1
        return lang[1]

    def serialize_hook(self, treestring: str) -> bytes:
        pos = 0
        out = []
        while pos >= 0:
            nextpos = treestring.find("<", pos)
            out.append(treestring[pos:nextpos].translate(OUTPUT_TRANS))
            pos = nextpos
            nextpos = treestring.find(">", pos)
            out.append(treestring[pos:nextpos])
            pos = nextpos
        out.append(treestring[pos:])
        return super().serialize_hook("".join(out))

    def serialize(self, out):
        """Write the XML document to a file."""
        root = self.document.getroot()
        for e in root.xpath(
            "//*[not(./node()) and not(text()) and not(name() = 'location')]"
        ):
            e.text = ""
        super().serialize(out)



from translate.storage import pocommon


def test_roundtrip_quote_plus():
    """Test that what we put in is what we get out."""

    def roundtrip_quote_plus(text, quoted):
        quote = pocommon.quote_plus(text)
        assert quote == quoted
        unquote = pocommon.unquote_plus(quoted)
        assert unquote == text

    roundtrip_quote_plus("abc", "abc")
    roundtrip_quote_plus("key space", "key+space")
    roundtrip_quote_plus("key ḓey", "key+%E1%B8%93ey")
    roundtrip_quote_plus(
        "path/file.c(2):3,path space/file.h:4", "path/file.c(2):3,path+space/file.h:4"
    )
    roundtrip_quote_plus("[uid1@example.com]SUMMARY", "[uid1@example.com]SUMMARY")



#
# Copyright 2006-2010 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""module for handling TBX glossary files."""

from lxml import etree

from translate.misc.xml_helpers import (
    getXMLspace,
    safely_set_text,
    setXMLlang,
)
from translate.storage import lisa


class tbxunit(lisa.LISAunit):
    """
    A single term in the TBX file.  Provisional work is done to make several
    languages possible.
    """

    rootNode = "termEntry"
    languageNode = "langSet"
    textNode = "term"

    def createlanguageNode(self, lang, text, purpose):
        """Returns a langset xml Element setup with given parameters."""
        langset = etree.Element(self.languageNode)
        setXMLlang(langset, lang)
        tig = etree.SubElement(langset, "tig")  # or ntig with termGrp inside
        term = etree.SubElement(tig, self.textNode)
        # probably not what we want:
        # lisa.setXMLspace(term, "preserve")
        safely_set_text(term, text)
        return langset

    def getid(self):
        # The id attribute is optional
        return self.xmlelement.get("id") or self.source

    def setid(self, value):
        return self.xmlelement.set("id", value)

    def _get_origin_element(self, origin: str):
        if origin == "pos":
            return self.namespaced("termNote")
        if origin == "definition":
            return self.namespaced("descrip")
        return self.namespaced("note")

    def removenotes(self, origin=None):
        """Remove all the translator notes."""
        notes = self.xmlelement.iterdescendants(self._get_origin_element(origin))
        for note in notes:
            self.xmlelement.remove(note)

    def addnote(self, text, origin=None, position="append"):
        """Add a note specifically in a "note" tag."""
        if position != "append":
            self.removenotes(origin=origin)

        if text:
            text = text.strip()
        if not text:
            return
        note = etree.SubElement(self.xmlelement, self._get_origin_element(origin))
        safely_set_text(note, text)
        if origin and origin not in {"pos", "definition"}:
            note.set("from", origin)

    def _getnotelist(self, origin=None):
        """
        Returns the text from notes matching ``origin`` or all notes.

        :param origin: The origin of the note (or note type)
        :type origin: String
        :return: The text from notes matching ``origin``
        :rtype: List
        """
        note_nodes = self.xmlelement.iterdescendants(self._get_origin_element(origin))
        # TODO: consider using xpath to construct initial_list directly
        # or to simply get the correct text from the outset (just remember to
        # check for duplication.
        initial_list = [
            lisa.getText(note, getXMLspace(self.xmlelement, self._default_xml_space))
            for note in note_nodes
            if origin in {"pos", "definition", None} or note.get("from") == origin
        ]

        # Remove duplicate entries from list:
        dictset = {}
        return [
            dictset.setdefault(note, note)
            for note in initial_list
            if note not in dictset
        ]

    def getnotes(self, origin=None):
        return "\n".join(self._getnotelist(origin=origin))


class tbxfile(lisa.LISAfile):
    """Class representing a TBX file store."""

    UnitClass = tbxunit
    Name = "TBX Glossary"
    Mimetypes = ["application/x-tbx"]
    Extensions = ["tbx"]
    rootNode = "martif"
    bodyNode = "body"
    XMLskeleton = """<?xml version="1.0"?>
<!DOCTYPE martif PUBLIC "ISO 12200:1999A//DTD MARTIF core (DXFcdV04)//EN" "TBXcdv04.dtd">
<martif type="TBX">
<martifHeader>
<fileDesc>
<sourceDesc><p>Translate Toolkit</p></sourceDesc>
</fileDesc>
</martifHeader>
<text><body></body></text>
</martif>"""
    XMLindent = {"indent": "    ", "toplevel": False}

    def addheader(self):
        """Initialise headers with TBX specific things."""
        setXMLlang(self.document.getroot(), self.sourcelanguage)



from io import BytesIO

from pytest import mark

from translate.convert import pot2po
from translate.storage import po

from . import test_convert


class TestPOT2PO:
    @staticmethod
    def convertpot(potsource, posource=None):
        """Helper that converts pot source to po source without requiring files."""
        potfile = BytesIO(potsource.encode())
        pofile = BytesIO(posource.encode()) if posource else None
        pooutfile = BytesIO()
        pot2po.convertpot(potfile, pooutfile, pofile)
        pooutfile.seek(0)
        return po.pofile(pooutfile.read())

    @staticmethod
    def singleunit(pofile):
        """Checks that the pofile contains a single non-header unit, and returns it."""
        assert len(pofile.units) == 2
        assert pofile.units[0].isheader()
        print(pofile.units[1])
        return pofile.units[1]

    def test_convertpot_blank(self):
        """Checks that the convertpot function is working for a simple file initialisation."""
        potsource = f"""#: simple.label{po.lsep}simple.accesskey\nmsgid "A &hard coded newline.\\n"\nmsgstr ""\n"""
        newpo = self.convertpot(potsource)
        assert str(self.singleunit(newpo)) == potsource

    def test_convertpot_blank_plurals(self):
        """Checks that the convertpot function is working for initialising plurals correctly."""
        potsource = r"""msgid ""
msgstr""

msgid "%d manual"
msgid_plural "%d manuals"
msgstr[0] ""
msgstr[1] ""
"""
        posource = r"""msgid ""
msgstr""
"Plural-Forms: nplurals=1; plural=0;\n"
"""

        poexpected = r"""msgid "%d manual"
msgid_plural "%d manuals"
msgstr[0] ""
"""
        newpo = self.convertpot(potsource, posource)
        assert str(self.singleunit(newpo)) == poexpected

    def test_merging_simple(self):
        """Checks that the convertpot function is working for a simple merge."""
        potsource = f"""#: simple.label{po.lsep}simple.accesskey\nmsgid "A &hard coded newline.\\n"\nmsgstr ""\n"""
        posource = f"""#: simple.label{po.lsep}simple.accesskey\nmsgid "A &hard coded newline.\\n"\nmsgstr "&Hart gekoeerde nuwe lyne\\n"\n"""
        newpo = self.convertpot(potsource, posource)
        assert str(self.singleunit(newpo)) == posource

    def test_merging_messages_marked_fuzzy(self):
        """Test that when we merge PO files with a fuzzy message that it remains fuzzy."""
        potsource = f"""#: simple.label{po.lsep}simple.accesskey\nmsgid "A &hard coded newline.\\n"\nmsgstr ""\n"""
        posource = f"""#: simple.label{po.lsep}simple.accesskey\n#, fuzzy\nmsgid "A &hard coded newline.\\n"\nmsgstr "&Hart gekoeerde nuwe lyne\\n"\n"""
        newpo = self.convertpot(potsource, posource)
        assert str(self.singleunit(newpo)) == posource

    def test_merging_plurals_with_fuzzy_matching(self):
        """Test that when we merge PO files with a fuzzy message that it remains fuzzy."""
        potsource = r"""#: file.cpp:2
msgid "%d manual"
msgid_plural "%d manuals"
msgstr[0] ""
msgstr[1] ""
"""
        posource = r"""#: file.cpp:3
#, fuzzy
msgid "%d manual"
msgid_plural "%d manuals"
msgstr[0] "%d handleiding."
msgstr[1] "%d handleidings."
"""
        # The #: comment and msgid's are different between the pot and the po
        poexpected = r"""#: file.cpp:2
#, fuzzy
msgid "%d manual"
msgid_plural "%d manuals"
msgstr[0] "%d handleiding."
msgstr[1] "%d handleidings."
"""
        newpo = self.convertpot(potsource, posource)
        assert str(self.singleunit(newpo)) == poexpected

    @mark.xfail(reason="Not implemented - review if this is even correct")
    def test_merging_msgid_change(self):
        """Tests that if the msgid changes but the location stays the same that we merge."""
        potsource = """#: simple.label\n#: simple.accesskey\nmsgid "Its &hard coding a newline.\\n"\nmsgstr ""\n"""
        posource = """#: simple.label\n#: simple.accesskey\nmsgid "A &hard coded newline.\\n"\nmsgstr "&Hart gekoeerde nuwe lyne\\n"\n"""
        poexpected = """#: simple.label\n#: simple.accesskey\n#, fuzzy\nmsgid "Its &hard coding a newline.\\n"\nmsgstr "&Hart gekoeerde nuwe lyne\\n"\n"""
        newpo = self.convertpot(potsource, posource)
        print(newpo)
        assert str(self.singleunit(newpo)) == poexpected

    def test_merging_location_change(self):
        """Tests that if the location changes but the msgid stays the same that we merge."""
        potsource = f"""#: new_simple.label{po.lsep}new_simple.accesskey\nmsgid "A &hard coded newline.\\n"\nmsgstr ""\n"""
        posource = f"""#: simple.label{po.lsep}simple.accesskey\nmsgid "A &hard coded newline.\\n"\nmsgstr "&Hart gekoeerde nuwe lyne\\n"\n"""
        poexpected = f"""#: new_simple.label{po.lsep}new_simple.accesskey\nmsgid "A &hard coded newline.\\n"\nmsgstr "&Hart gekoeerde nuwe lyne\\n"\n"""
        newpo = self.convertpot(potsource, posource)
        print(newpo)
        assert str(self.singleunit(newpo)) == poexpected

    def test_merging_location_and_whitespace_change(self):
        """
        Test that even if the location changes that if the msgid
        only has whitespace changes we can still merge.
        """
        potsource = f"""#: singlespace.label{po.lsep}singlespace.accesskey\nmsgid "&We have spaces"\nmsgstr ""\n"""
        posource = f"""#: doublespace.label{po.lsep}doublespace.accesskey\nmsgid "&We  have  spaces"\nmsgstr "&One  het  spasies"\n"""
        poexpected = f"""#: singlespace.label{po.lsep}singlespace.accesskey\n#, fuzzy\nmsgid "&We have spaces"\nmsgstr "&One  het  spasies"\n"""
        newpo = self.convertpot(potsource, posource)
        print(newpo)
        assert str(self.singleunit(newpo)) == poexpected

    def test_merging_location_ambiguous_with_disambiguous(self):
        """
        Test that when we have a PO in ambiguous (Gettext form) and merge with
        disamabiguous (KDE comment form) that we don't duplicate the
        location #: comments.
        """
        potsource = (
            """#: location.c:1\nmsgid ""\n"_: location.c:1\\n"\n"Source"\nmsgstr ""\n\n"""
            """#: location.c:10\nmsgid ""\n"_: location.c:10\\n"\n"Source"\nmsgstr ""\n"""
        )
        posource = (
            """#: location.c:1\n#: location.c:10\nmsgid "Source"\nmsgstr "Target"\n\n"""
        )
        poexpected1 = """#: location.c:1\n#, fuzzy\nmsgid ""\n"_: location.c:1\\n"\n"Source"\nmsgstr "Target"\n"""
        poexpected2 = """#: location.c:10\n#, fuzzy\nmsgid ""\n"_: location.c:10\\n"\n"Source"\nmsgstr "Target"\n"""
        newpo = self.convertpot(potsource, posource)
        print("Expected:\n", poexpected1, "Actual:\n", newpo.units[1])
        assert str(newpo.units[1]) == poexpected1
        assert str(newpo.units[2]) == poexpected2

    @mark.xfail(reason="Not Implemented - needs review")
    def test_merging_accelerator_changes(self):
        """Test that a change in the accelerator localtion still allows merging."""
        potsource = """#: someline.c\nmsgid "A&bout"\nmsgstr ""\n"""
        posource = """#: someline.c\nmsgid "&About"\nmsgstr "&Info"\n"""
        poexpected = """#: someline.c\nmsgid "A&bout"\nmsgstr "&Info"\n"""
        newpo = self.convertpot(potsource, posource)
        print(newpo)
        assert str(self.singleunit(newpo)) == poexpected

    @mark.xfail(reason="Not Implemented - review if this is even correct")
    def test_lines_cut_differently(self):
        """Checks that the correct formatting is preserved when pot an po lines differ."""
        potsource = (
            """#: simple.label\nmsgid "Line split "\n"differently"\nmsgstr ""\n"""
        )
        posource = """#: simple.label\nmsgid "Line"\n" split differently"\nmsgstr "Lyne verskillend gesny"\n"""
        newpo = self.convertpot(potsource, posource)
        newpounit = self.singleunit(newpo)
        assert str(newpounit) == posource

    def test_merging_automatic_comments_dont_duplicate(self):
        """Ensure that we can merge #. comments correctly."""
        potsource = """#. Row 35\nmsgid "&About"\nmsgstr ""\n"""
        posource = """#. Row 35\nmsgid "&About"\nmsgstr "&Info"\n"""
        newpo = self.convertpot(potsource, posource)
        newpounit = self.singleunit(newpo)
        assert str(newpounit) == posource

    def test_merging_automatic_comments_new_overides_old(self):
        """Ensure that new #. comments override the old comments."""
        potsource = """#. new comment\n#: someline.c\nmsgid "&About"\nmsgstr ""\n"""
        posource = """#. old comment\n#: someline.c\nmsgid "&About"\nmsgstr "&Info"\n"""
        poexpected = (
            """#. new comment\n#: someline.c\nmsgid "&About"\nmsgstr "&Info"\n"""
        )
        newpo = self.convertpot(potsource, posource)
        newpounit = self.singleunit(newpo)
        assert str(newpounit) == poexpected

    def test_merging_comments_with_blank_comment_lines(self):
        """Test that when we merge a comment that has a blank line we keep the blank line."""
        potsource = """#: someline.c\nmsgid "About"\nmsgstr ""\n"""
        posource = """# comment1\n#\n# comment2\n#: someline.c\nmsgid "About"\nmsgstr "Omtrent"\n"""
        poexpected = posource
        newpo = self.convertpot(potsource, posource)
        newpounit = self.singleunit(newpo)
        assert str(newpounit) == poexpected

    def test_empty_commentlines(self):
        potsource = """#: paneSecurity.title
msgid "Security"
msgstr ""
"""
        posource = """# - Contributor(s):
# -
# - Alternatively, the
# -
#: paneSecurity.title
msgid "Security"
msgstr "Sekuriteit"
"""
        poexpected = posource
        newpo = self.convertpot(potsource, posource)
        newpounit = self.singleunit(newpo)
        print("expected")
        print(poexpected)
        print("got:")
        print(str(newpounit))
        assert str(newpounit) == poexpected

    def test_merging_msgidcomments(self):
        """Ensure that we can merge msgidcomments messages."""
        potsource = r"""#: window.width
msgid ""
"_: Do not translate this.\n"
"36em"
msgstr ""
"""
        posource = r"""#: window.width
msgid ""
"_: Do not translate this.\n"
"36em"
msgstr "36em"
"""
        newpo = self.convertpot(potsource, posource)
        newpounit = self.singleunit(newpo)
        assert str(newpounit) == posource

    def test_merging_msgid_with_msgidcomment(self):
        """Test that we can merge an otherwise identical string that has a different msgid."""
        potsource = r"""#: pref.certs.title
msgid ""
"_: pref.certs.title\n"
"Certificates"
msgstr ""

#: certs.label
msgid ""
"_: certs.label\n"
"Certificates"
msgstr ""
"""
        posource = r"""#: pref.certs.title
msgid ""
"_: pref.certs.title\n"
"Certificates"
msgstr ""

#: certs.label
msgid ""
"_: certs.label\n"
"Certificates"
msgstr "Sertifikate"
"""
        expected = r"""#: pref.certs.title
#, fuzzy
msgid ""
"_: pref.certs.title\n"
"Certificates"
msgstr "Sertifikate"
"""
        newpo = self.convertpot(potsource, posource)
        newpounit = newpo.units[1]
        assert str(newpounit) == expected

    def test_merging_plurals(self):
        """Ensure that we can merge plural messages."""
        potsource = """msgid "One"\nmsgid_plural "Two"\nmsgstr[0] ""\nmsgstr[1] ""\n"""
        posource = """msgid "One"\nmsgid_plural "Two"\nmsgstr[0] "Een"\nmsgstr[1] "Twee"\nmsgstr[2] "Drie"\n"""
        newpo = self.convertpot(potsource, posource)
        print(newpo)
        newpounit = self.singleunit(newpo)
        assert str(newpounit) == posource

    def test_merging_obsoleting_messages(self):
        """Check that we obsolete messages no longer present in the new file."""
        # add emtpy msgid line to help factory identify format
        potsource = 'msgid ""\nmsgstr ""\n'
        posource = '# Some comment\n#. Extracted comment\n#: obsoleteme:10\nmsgid "One"\nmsgstr "Een"\n'
        expected = '# Some comment\n#~ msgid "One"\n#~ msgstr "Een"\n'
        newpo = self.convertpot(potsource, posource)
        print(bytes(newpo))
        newpounit = self.singleunit(newpo)
        assert str(newpounit) == expected

    def test_not_obsoleting_empty_messages(self):
        """Check that we don't obsolete (and keep) untranslated messages."""
        # add emtpy msgid line to help factory identify format
        potsource = 'msgid ""\nmsgstr ""\n'
        posource = '#: obsoleteme:10\nmsgid "One"\nmsgstr ""\n'
        newpo = self.convertpot(potsource, posource)
        print(bytes(newpo))
        # We should only have the header
        assert len(newpo.units) == 1

    def test_merging_new_before_obsolete(self):
        """Test to check that we place new blank message before obsolete messages."""
        potsource = """#: newline.c\nmsgid "&About"\nmsgstr ""\n"""
        posource = """#~ msgid "Old"\n#~ msgstr "Oud"\n"""
        newpo = self.convertpot(potsource, posource)
        assert len(newpo.units) == 3
        assert newpo.units[0].isheader()
        assert newpo.units[2].isobsolete()
        assert str(newpo.units[1]) == potsource
        assert str(newpo.units[2]) == posource

        # Now test with real units present in posource
        posource2 = """msgid "Old"\nmsgstr "Oud"\n"""
        newpo = self.convertpot(potsource, posource2)
        assert len(newpo.units) == 3
        assert newpo.units[0].isheader()
        assert newpo.units[2].isobsolete()
        assert str(newpo.units[1]) == potsource
        assert str(newpo.units[2]) == posource

    def test_merging_resurect_obsolete_messages(self):
        """Check that we can reuse old obsolete messages if the message comes back."""
        potsource = """#: resurect.c\nmsgid "&About"\nmsgstr ""\n"""
        posource = """#~ msgid "&About"\n#~ msgstr "&Omtrent"\n"""
        expected = """#: resurect.c\nmsgid "&About"\nmsgstr "&Omtrent"\n"""
        newpo = self.convertpot(potsource, posource)
        print(newpo)
        assert len(newpo.units) == 2
        assert newpo.units[0].isheader()
        newpounit = self.singleunit(newpo)
        assert str(newpounit) == expected

    def test_merging_resurect_obsolete_messages_into_msgidcomment(self):
        """Check that we can reuse old obsolete messages even if the recipient has a msgidcomment."""
        potsource = (
            """#: resurect1.c\nmsgid "About"\nmsgstr ""\n\n"""
            """#: resurect2.c\nmsgid ""\n"_: resurect2.c\\n"\n"About"\nmsgstr ""\n"""
        )
        posource = """#~ msgid "About"\n#~ msgstr "Omtrent"\n"""
        expected1 = """#: resurect1.c\nmsgid "About"\nmsgstr "Omtrent"\n"""
        expected2 = """#: resurect2.c\n#, fuzzy\nmsgid ""\n"_: resurect2.c\\n"\n"About"\nmsgstr "Omtrent"\n"""
        newpo = self.convertpot(potsource, posource)
        print(newpo)
        assert len(newpo.units) == 3
        assert newpo.units[0].isheader()
        assert str(newpo.units[1]) == expected1
        assert str(newpo.units[2]) == expected2

    def test_header_initialisation(self):
        """Test to check that we initialise the header correctly."""
        potsource = r"""#, fuzzy
msgid ""
msgstr ""
"Project-Id-Version: PACKAGE VERSION\n"
"Report-Msgid-Bugs-To: new@example.com\n"
"POT-Creation-Date: 2006-11-11 11:11+0000\n"
"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\n"
"Last-Translator: FULL NAME <EMAIL@ADDRESS>\n"
"Language-Team: LANGUAGE <LL@li.org>\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"Plural-Forms: nplurals=INTEGER; plural=EXPRESSION;\n"
"X-Generator: Translate Toolkit 0.10rc2\n"
"""
        posource = r"""msgid ""
msgstr ""
"Project-Id-Version: Pootle 0.10\n"
"Report-Msgid-Bugs-To: old@example.com\n"
"POT-Creation-Date: 2006-01-01 01:01+0100\n"
"PO-Revision-Date: 2006-09-09 09:09+0900\n"
"Last-Translator: Joe Translate <joe@example.com>\n"
"Language-Team: Pig Latin <piglatin@example.com>\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"Plural-Forms: nplurals=2; plural=(n != 1);\n"
"X-Generator: Translate Toolkit 0.9\n"
"""
        expected = r"""msgid ""
msgstr ""
"Project-Id-Version: Pootle 0.10\n"
"Report-Msgid-Bugs-To: new@example.com\n"
"POT-Creation-Date: 2006-11-11 11:11+0000\n"
"PO-Revision-Date: 2006-09-09 09:09+0900\n"
"Last-Translator: Joe Translate <joe@example.com>\n"
"Language-Team: Pig Latin <piglatin@example.com>\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"Plural-Forms: nplurals=2; plural=(n != 1);\n"
"X-Generator: Translate Toolkit 0.10rc2\n"
"""
        newpo = self.convertpot(potsource, posource)
        print(f"Output Header:\n{newpo}")
        print(f"Expected Header:\n{expected}")
        assert bytes(newpo).decode("utf-8") == expected

    def test_merging_comments(self):
        """Test that we can merge comments correctly."""
        potsource = """#. Don't do it!\n#: file.py:1\nmsgid "One"\nmsgstr ""\n"""
        posource = """#. Don't do it!\n#: file.py:2\nmsgid "One"\nmsgstr "Een"\n"""
        poexpected = """#. Don't do it!\n#: file.py:1\nmsgid "One"\nmsgstr "Een"\n"""
        newpo = self.convertpot(potsource, posource)
        print(newpo)
        newpounit = self.singleunit(newpo)
        assert str(newpounit) == poexpected

    def test_merging_typecomments(self):
        """Test that we can merge with typecomments."""
        potsource = """#: file.c:1\n#, c-format\nmsgid "%d pipes"\nmsgstr ""\n"""
        posource = """#: file.c:2\nmsgid "%d pipes"\nmsgstr "%d pype"\n"""
        poexpected = (
            """#: file.c:1\n#, c-format\nmsgid "%d pipes"\nmsgstr "%d pype"\n"""
        )
        newpo = self.convertpot(potsource, posource)
        newpounit = self.singleunit(newpo)
        print(newpounit)
        assert str(newpounit) == poexpected

        potsource = """#: file.c:1\n#, c-format\nmsgid "%d computers"\nmsgstr ""\n"""
        posource = """#: file.c:2\n#, c-format\nmsgid "%s computers "\nmsgstr "%s-rekenaars"\n"""
        poexpected = """#: file.c:1\n#, fuzzy, c-format\nmsgid "%d computers"\nmsgstr "%s-rekenaars"\n"""
        newpo = self.convertpot(potsource, posource)
        newpounit = self.singleunit(newpo)
        assert newpounit.isfuzzy()
        assert newpounit.hastypecomment("c-format")

    def test_msgctxt(self):
        """Test that msgctxt is migrated correctly."""
        potsource = """
#: something.h:5
msgctxt "context1"
msgid "text"
msgstr ""

#: something.h:6
msgctxt "context2"
msgid "text"
msgstr ""
"""
        posource = """
#: something.h:3
msgctxt "context0"
msgid "text"
msgstr "teks"

#: something.h:4
msgctxt "context1"
msgid "text"
msgstr "sms"
"""
        poexpected = """
#: something.h:5
msgctxt "context1"
msgid "text"
msgstr "sms"

#: something.h:6
#, fuzzy
msgctxt "context2"
msgid "text"
msgstr "teks"
"""
        newpo = self.convertpot(potsource, posource)
        print(newpo)
        assert poexpected in bytes(newpo).decode("utf-8")

    def test_msgctxt_multiline(self):
        """Test multiline msgctxt fields."""
        pot_source = r"""#. |1MV
#: ActionTe.ulf
msgctxt ""
"ActionTe.ulf\n"
"OOO_ACTIONTEXT_21\n"
"LngText.text"
msgid "Computing space requirements"
msgstr ""
"""

        po_source = r"""#. |1MV
#: ActionTe.ulf
msgctxt ""
"ActionTe.ulf\n"
"OOO_ACTIONTEXT_21\n"
"LngText.text"
msgid "Computing space requirements"
msgstr "A szükséges lemezterület kiszámítása"
"""

        new_po = self.convertpot(pot_source, po_source)
        assert new_po.units[0].isheader()
        unit = new_po.units[1]
        assert not unit.isfuzzy()
        assert po_source == str(unit)

    def test_msgid_merge_on_location(self):
        """Tests that unit merges rely on location-based matching."""
        pot_source = r"""
msgid ""
msgstr ""
"X-Merge-On: location\n"

#. $:am
#: ActionTe.ulf
msgctxt ""
"ActionTe.ulf\n"
"OOO_ACTIONTEXT_11\n"
"LngText.text"
msgid "Computing space requirements"
msgstr ""

#. NjJ3
#: ActionTe.ulf
msgctxt ""
"ActionTe.ulf\n"
"OOO_ACTIONTEXT_12\n"
"LngText.text"
msgid "Computing space requirements"
msgstr ""
"""

        po_source = r"""
#. $:am
#: ActionTe.ulf
msgctxt ""
"ActionTe.ulf\n"
"OOO_ACTIONTEXT_11\n"
"LngText.text"
msgid "Computing space requirements"
msgstr "A szükséges lemezterület kiszámítása"

#. NjJ3
#: ActionTe.ulf
msgctxt ""
"ActionTe.ulf\n"
"OOO_ACTIONTEXT_12\n"
"LngText.text"
msgid "Computing space requirements"
msgstr "A szükséges lemezterület kiszámítása"
"""

        new_po = self.convertpot(pot_source, po_source)

        assert len(new_po.units) == 3
        assert new_po.units[0].isheader()

        assert not new_po.units[1].isfuzzy()
        assert new_po.units[2].isfuzzy()

    def test_msgid_merge_on_id(self):
        """Tests that unit merges rely on location-based matching."""
        pot_source = r"""
msgid ""
msgstr ""
"X-Merge-On: id\n"

#. $:am
#: ActionTe.ulf
msgctxt ""
"ActionTe.ulf\n"
"OOO_ACTIONTEXT_11\n"
"LngText.text"
msgid "Computing space requirements"
msgstr ""

#. NjJ3
#: ActionTe.ulf
msgctxt ""
"ActionTe.ulf\n"
"OOO_ACTIONTEXT_12\n"
"LngText.text"
msgid "Computing space requirements"
msgstr ""
"""
        pot_source_noheader = r"""
#. $:am
#: ActionTe.ulf
msgctxt ""
"ActionTe.ulf\n"
"OOO_ACTIONTEXT_11\n"
"LngText.text"
msgid "Computing space requirements"
msgstr ""

#. NjJ3
#: ActionTe.ulf
msgctxt ""
"ActionTe.ulf\n"
"OOO_ACTIONTEXT_12\n"
"LngText.text"
msgid "Computing space requirements"
msgstr ""
"""

        po_source = r"""
#. $:am
#: ActionTe.ulf
msgctxt ""
"ActionTe.ulf\n"
"OOO_ACTIONTEXT_11\n"
"LngText.text"
msgid "Computing space requirements"
msgstr "A szükséges lemezterület kiszámítása"

#. NjJ3
#: ActionTe.ulf
msgctxt ""
"ActionTe.ulf\n"
"OOO_ACTIONTEXT_12\n"
"LngText.text"
msgid "Computing space requirements"
msgstr "A szükséges lemezterület kiszámítása"
"""

        for pot in (pot_source, pot_source_noheader):
            new_po = self.convertpot(pot, po_source)

            assert len(new_po.units) == 3
            assert new_po.units[0].isheader()

            assert not new_po.units[1].isfuzzy()
            assert not new_po.units[2].isfuzzy()

    def test_empty_msgid(self):
        """Test that we handle empty msgids correctly."""
        # TODO: this test will fail if we don't have the gettext location
        # comment in the pot file
        potsource = '#: file:1\nmsgctxt "bla"\nmsgid ""\nmsgstr ""\n'
        posource = r"""
msgid ""
"Project-Id-Version: Pootle 0.10\n"
msgstr ""

msgctxt "bla"
msgid ""
msgstr "trans"
"""
        newpo = self.convertpot(potsource, posource)
        print(newpo)
        assert len(newpo.units) == 2
        assert newpo.units[0].isheader()
        unit = newpo.units[1]
        assert unit.source == ""
        assert unit.getid() == "bla\04"
        assert unit.target == "trans"
        assert not unit.isfuzzy()

    def test_migrate_msgidcomment_to_msgctxt(self):
        """
        Test that we migrate correctly from msgidcomments to msgctxt.

        This is needed for our move away from using msgidcomments for mozilla.
        """
        potsource = r"""
msgid ""
msgstr ""
"X-Accelerator-Marker: &\n"
"X-Merge-On: location\n"


#: bla
msgctxt "bla"
msgid ""
msgstr ""
"""
        posource = r"""
msgid ""
msgstr ""
"Project-Id-Version: Pootle 0.10\n"
"X-Accelerator-Marker: &\n"


#: bla
msgid ""
"_: bla\n"
msgstr "trans"
"""
        newpo = self.convertpot(potsource, posource)
        print(newpo)
        assert len(newpo.units) == 2
        assert newpo.units[0].isheader()
        unit = newpo.units[1]
        assert unit.source == ""
        assert unit.getid() == "bla\04"
        assert unit.target == "trans"
        assert not unit.isfuzzy()

    def test_obsolete_msgctxt(self):
        """Test that obsolete units' msgctxt is preserved."""
        potsource = 'msgctxt "newContext"\nmsgid "First unit"\nmsgstr ""'
        posource = """
msgctxt "newContext"
msgid "First unit"
msgstr "Eerste eenheid"

#~ msgctxt "context"
#~ msgid "Old unit"
#~ msgstr "Ou eenheid1"

#~ msgctxt "context2"
#~ msgid "Old unit"
#~ msgstr "Ou eenheid2"

#~ msgid "Old unit"
#~ msgstr "Ou eenheid3"
"""
        newpo = self.convertpot(potsource, posource)
        print(newpo)
        assert len(newpo.units) == 5
        assert newpo.units[1].getcontext() == "newContext"
        # Search in unit string, because obsolete units can't return a context
        assert 'msgctxt "context"' in str(newpo.units[2])
        assert 'msgctxt "context2"' in str(newpo.units[3])

    def test_small_strings(self):
        """
        Test that units with small source strings are not incorrectly
        populated by means of fuzzy matching.
        """
        potsource = r"""#, fuzzy
msgid ""
msgstr ""
"Project-Id-Version: PACKAGE VERSION\n"
"Report-Msgid-Bugs-To: new@example.com\n"
"POT-Creation-Date: 2006-11-11 11:11+0000\n"
"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\n"
"Last-Translator: FULL NAME <EMAIL@ADDRESS>\n"
"Language-Team: LANGUAGE <LL@li.org>\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"Plural-Forms: nplurals=INTEGER; plural=EXPRESSION;\n"
"X-Accelerator-Marker: &\n"
"X-Generator: Translate Toolkit 0.10rc2\n"
"X-Merge-On: location\n"

#: new_disassociated_mozilla_accesskey
msgid "R"
msgstr ""
"""
        posource = r"""msgid ""
msgstr ""
"Project-Id-Version: Pootle 0.10\n"
"Report-Msgid-Bugs-To: old@example.com\n"
"POT-Creation-Date: 2006-01-01 01:01+0100\n"
"PO-Revision-Date: 2006-09-09 09:09+0900\n"
"Last-Translator: Joe Translate <joe@example.com>\n"
"Language-Team: Pig Latin <piglatin@example.com>\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"Plural-Forms: nplurals=2; plural=(n != 1);\n"
"X-Generator: Translate Toolkit 0.9\n"
"X-Accelerator-Marker: &\n"

#: old_disassociated_mozilla_accesskey
msgid "R"
msgstr "S"
"""
        expected = r"""msgid ""
msgstr ""
"Project-Id-Version: Pootle 0.10\n"
"Report-Msgid-Bugs-To: new@example.com\n"
"POT-Creation-Date: 2006-11-11 11:11+0000\n"
"PO-Revision-Date: 2006-09-09 09:09+0900\n"
"Last-Translator: Joe Translate <joe@example.com>\n"
"Language-Team: Pig Latin <piglatin@example.com>\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"Plural-Forms: nplurals=2; plural=(n != 1);\n"
"X-Accelerator-Marker: &\n"
"X-Generator: Translate Toolkit 0.10rc2\n"
"X-Merge-On: location\n"

#: new_disassociated_mozilla_accesskey
msgid "R"
msgstr ""
"""
        newpo = self.convertpot(potsource, posource)
        print(f"Output:\n{newpo}")
        print(f"Expected:\n{expected}")
        assert bytes(newpo).decode("utf-8") == expected


class TestPOT2POCommand(test_convert.TestConvertCommand, TestPOT2PO):
    """Tests running actual pot2po commands on files."""

    convertmodule = pot2po

    expected_options = [
        "-t TEMPLATE, --template=TEMPLATE",
        "-P, --pot",
        "--tm",
        "-s MIN_SIMILARITY, --similarity=MIN_SIMILARITY",
        "--nofuzzymatching",
    ]



from translate.misc.multistring import multistring
from translate.storage import poxliff

from . import test_xliff


class TestPOXLIFFUnit(test_xliff.TestXLIFFUnit):
    UnitClass = poxliff.PoXliffUnit

    def test_plurals(self):
        """Tests that plurals are handled correctly."""
        unit = self.UnitClass(multistring(["Cow", "Cows"]))
        print(type(unit.source))
        print(repr(unit.source))
        assert isinstance(unit.source, multistring)
        assert unit.source.strings == ["Cow", "Cows"]
        assert unit.source == "Cow"

        unit.target = ["Koei", "Koeie"]
        assert isinstance(unit.target, multistring)
        assert unit.target.strings == ["Koei", "Koeie"]
        assert unit.target == "Koei"

        unit.target = ["Sk\u00ear", "Sk\u00eare"]
        assert isinstance(unit.target, multistring)
        assert unit.target.strings == ["Sk\u00ear", "Sk\u00eare"]
        assert unit.target.strings == ["Sk\u00ear", "Sk\u00eare"]
        assert unit.target == "Sk\u00ear"

    def test_ids(self):
        """Tests that ids are assigned correctly, especially for plurals."""
        unit = self.UnitClass("gras")
        assert not unit.getid()
        unit.setid("4")
        assert unit.getid() == "4"

        unit = self.UnitClass(multistring(["shoe", "shoes"]))
        assert not unit.getid()
        unit.setid("20")
        assert unit.getid() == "20"
        assert unit.units[1].getid() == "20[1]"

        unit.target = ["utshani", "uutshani", "uuutshani"]
        assert unit.getid() == "20"
        assert unit.units[1].getid() == "20[1]"


class TestPOXLIFFfile(test_xliff.TestXLIFFfile):
    StoreClass = poxliff.PoXliffFile
    xliffskeleton = """<?xml version="1.0" ?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
  <file original="filename.po" source-language="en-US" datatype="po">
    <body>
        %s
    </body>
  </file>
</xliff>"""

    def test_parse(self):
        minixlf = (
            self.xliffskeleton
            % """<group restype="x-gettext-plurals">
        <trans-unit id="1[0]" xml:space="preserve">
            <source>cow</source>
            <target>inkomo</target>
        </trans-unit>
        <trans-unit id="1[1]" xml:space="preserve">
            <source>cows</source>
            <target>iinkomo</target>
        </trans-unit>
</group>"""
        )
        xlifffile = self.StoreClass.parsestring(minixlf)
        assert len(xlifffile.units) == 1
        assert xlifffile.translate("cow") == "inkomo"
        assert xlifffile.units[0].source == "cow"
        assert xlifffile.units[0].source == multistring(["cow", "cows"])

    def test_parse_plural_alpha_id(self):
        minixlf = (
            self.xliffskeleton
            % """<group restype="x-gettext-plurals">
        <trans-unit id="test[0]" xml:space="preserve">
            <source>cow</source>
            <target>inkomo</target>
        </trans-unit>
        <trans-unit id="test[1]" xml:space="preserve">
            <source>cows</source>
            <target>iinkomo</target>
        </trans-unit>
</group>"""
        )
        xlifffile = self.StoreClass.parsestring(minixlf)
        assert len(xlifffile.units) == 1
        assert xlifffile.translate("cow") == "inkomo"
        assert xlifffile.units[0].source == "cow"
        assert xlifffile.units[0].source == multistring(["cow", "cows"])

    def test_notes(self):
        minixlf = (
            self.xliffskeleton
            % """<group restype="x-gettext-plurals">
        <trans-unit id="1[0]" xml:space="preserve">
            <source>cow</source>
            <target>inkomo</target>
<note from="po-translator">Zulu translation of program ABC</note>
<note from="developer">azoozoo come back!</note>
        </trans-unit>
        <trans-unit id="1[1]" xml:space="preserve">
            <source>cows</source>
            <target>iinkomo</target>
<note from="po-translator">Zulu translation of program ABC</note>
<note from="developer">azoozoo come back!</note>
        </trans-unit>
</group>"""
        )
        xlifffile = self.StoreClass.parsestring(minixlf)
        assert (
            xlifffile.units[0].getnotes()
            == "Zulu translation of program ABC\nazoozoo come back!"
        )
        assert xlifffile.units[0].getnotes("developer") == "azoozoo come back!"
        assert (
            xlifffile.units[0].getnotes("po-translator")
            == "Zulu translation of program ABC"
        )

    def test_plural(self):
        minixlf = (
            self.xliffskeleton
            % """
        <group id="1238108068" restype="x-gettext-plurals">
                <trans-unit id="1238108068[0]" xml:space="preserve">
                        <source>This field must contain at least {0,number} character</source>
                </trans-unit>
                <trans-unit id="1238108068[1]" xml:space="preserve">
                        <source>This field must contain at least {0,number} characters</source>
                </trans-unit>
        </group>
            """
        )
        xlifffile = self.StoreClass.parsestring(minixlf)
        assert len(xlifffile.units) == 1
        unit = xlifffile.units[0]
        assert unit.source.strings == [
            "This field must contain at least {0,number} character",
            "This field must contain at least {0,number} characters",
        ]
        assert unit.target == ["", ""]



from pytest import mark

from translate.lang import common


def test_characters():
    """Test the basic characters segmentation."""
    language = common.Common
    assert language.characters("") == []
    assert language.characters("Four") == ["F", "o", "u", "r"]
    assert language.characters("A B") == ["A", " ", "B"]
    # Spaces are compacted, source has 2 returned has only one
    assert language.characters("A  B") == ["A", " ", "B"]


def test_words():
    """Tests basic functionality of word segmentation."""
    language = common.Common
    words = language.words("")
    assert words == []

    words = language.words("test sentence.")
    assert words == ["test", "sentence"]

    words = language.words("This is a weird test .")
    assert words == ["This", "is", "a", "weird", "test"]

    words = language.words("Don't send e-mail!")
    assert words == ["Don't", "send", "e-mail"]

    words = language.words("Don’t send e-mail!")
    assert words == ["Don’t", "send", "e-mail"]


@mark.xfail(
    reason="ZWS is not considered a space in Python 2.6+. Khmer "
    "should extend words() to include \\u200b in addition to "
    "other word breakers."
)
def test_word_khmer():
    language = common.Common
    # Let's test Khmer with zero width space (\u200b)
    words = language.words("ផ្ដល់\u200bយោបល់")
    print("ផ្ដល់\u200bយោបល់")
    print(language.words("ផ្ដល់<200b>យោបល់"))
    print(["ផ្ដល់", "យោបល់"])
    assert words == ["ផ្ដល់", "យោបល់"]


def test_sentences():
    """Tests basic functionality of sentence segmentation."""
    language = common.Common
    # Check that we correctly handle an empty string:
    sentences = language.sentences("")

    sentences = language.sentences("This is a sentence.")
    assert sentences == ["This is a sentence."]
    sentences = language.sentences("This is a sentence")
    assert sentences == ["This is a sentence"]
    sentences = language.sentences("This is a sentence. Another one.")
    assert sentences == ["This is a sentence.", "Another one."]
    sentences = language.sentences("This is a sentence. Another one. Bla.")
    assert sentences == ["This is a sentence.", "Another one.", "Bla."]
    sentences = language.sentences("This is a sentence.Not another one.")
    assert sentences == ["This is a sentence.Not another one."]
    sentences = language.sentences("Exclamation! Really? No...")
    assert sentences == ["Exclamation!", "Really?", "No..."]
    sentences = language.sentences("Four i.e. 1+3. See?")
    assert sentences == ["Four i.e. 1+3.", "See?"]
    sentences = language.sentences("Apples, bananas, etc. are nice.")
    assert sentences == ["Apples, bananas, etc. are nice."]
    sentences = language.sentences("Apples, bananas, etc.\nNext part")
    assert sentences == ["Apples, bananas, etc.", "Next part"]
    sentences = language.sentences(
        "No font for displaying text in encoding '%s' found,\nbut an alternative encoding '%s' is available.\nDo you want to use this encoding (otherwise you will have to choose another one)?"
    )
    assert sentences == [
        "No font for displaying text in encoding '%s' found,\nbut an alternative encoding '%s' is available.",
        "Do you want to use this encoding (otherwise you will have to choose another one)?",
    ]
    # Test that a newline at the end won't confuse us
    sentences = language.sentences("The first sentence. The second sentence.\n")
    assert sentences == ["The first sentence.", "The second sentence."]
    sentences = language.sentences("P.O. box")
    assert sentences == ["P.O. box"]
    sentences = language.sentences("Doen dit d.m.v. koeie.")
    assert sentences == ["Doen dit d.m.v. koeie."]


def test_capsstart():
    """Tests for basic sane behaviour in startcaps()."""
    language = common.Common
    assert language.capsstart("Open cow file")
    assert language.capsstart("'Open' cow file")
    assert not language.capsstart("open cow file")
    assert not language.capsstart(":")
    assert not language.capsstart("")


def test_numstart():
    """Tests for basic sane behaviour in startcaps()."""
    language = common.Common
    assert language.numstart("360 degress")
    assert language.numstart("3D file")
    assert not language.numstart("Open 360 degrees")
    assert not language.numstart(":")
    assert not language.numstart("")


def test_punctranslate():
    """Test the basic punctranslate function."""
    language = common.Common
    assert language.punctranslate("A...") != "A…"
    language.puncdict = {"...": "…"}
    assert language.punctranslate("A...") == "A…"


def test_length_difference():
    """Test the heuristics of the length difference function."""
    # Expansion with no code
    assert common.Common.length_difference(10) == 6
    assert common.Common.length_difference(100) == 15
    assert common.Common.length_difference(300) == 35


def test_alter_length():
    """Test that we create the correct length by adding or removing characters."""
    assert common.Common.alter_length("One two three") == "One twOne two three"



#
# Copyright 2004-2010 Zuza Software Foundation
#
# This file is part of translate.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Convert template files (like .pot or template .xlf files) to translation
files, preserving existing translations.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/pot2po.html
for examples and usage instructions.
"""

from translate.misc.multistring import multistring
from translate.search import match
from translate.storage import catkeys, factory, poheader
from translate.tools import pretranslate


def convertpot(
    input_file,
    output_file,
    template_file,
    tm=None,
    min_similarity=75,
    fuzzymatching=True,
    classes=None,
    classes_str=None,
    **kwargs,
):
    """Main conversion function."""
    input_store = factory.getobject(
        input_file, classes=classes, classes_str=classes_str
    )
    try:
        temp_store = factory.getobject(input_file, classes_str=classes_str)
    except Exception:
        # StringIO and other file like objects will be closed after parsing
        temp_store = None

    template_store = None
    if template_file is not None:
        template_store = factory.getobject(template_file, classes_str=classes_str)

    output_store = convert_stores(
        input_store,
        template_store,
        temp_store,
        tm,
        min_similarity,
        fuzzymatching,
        **kwargs,
    )
    output_store.serialize(output_file)

    return 1


def convert_stores(
    input_store,
    template_store,
    temp_store=None,
    tm=None,
    min_similarity=75,
    fuzzymatching=True,
    **kwargs,
):
    """
    Actual conversion function, works on stores not files, returns
    a properly initialized pretranslated output store, with structure
    based on input_store, metadata based on template_store, migrates
    old translations from template_store and pretranslating from TM.
    """
    if temp_store is None:
        temp_store = input_store

    # Create fuzzy matchers to be used by pretranslate.pretranslate_unit
    matchers = []

    _prepare_merge(input_store, temp_store, template_store)
    if fuzzymatching:
        if template_store:
            matcher = match.matcher(
                template_store,
                max_candidates=1,
                min_similarity=min_similarity,
                max_length=3000,
                usefuzzy=True,
            )
            matcher.addpercentage = False
            matchers.append(matcher)
        if tm:
            matcher = pretranslate.memory(
                tm, max_candidates=1, min_similarity=min_similarity, max_length=1000
            )
            matcher.addpercentage = False
            matchers.append(matcher)

    # initialize store
    _store_pre_merge(input_store, temp_store, template_store)

    # Do matching
    for input_unit in temp_store.units:
        if input_unit.istranslatable():
            input_unit = pretranslate.pretranslate_unit(
                input_unit,
                template_store,
                matchers,
                mark_reused=True,
                merge_on=input_store.merge_on,
            )
            _unit_post_merge(input_unit, input_store, temp_store, template_store)

    # finalize store
    _store_post_merge(input_store, temp_store, template_store)

    return temp_store


##dispatchers
def _prepare_merge(input_store, output_store, template_store, **kwargs):
    """Prepare stores & TM matchers before merging."""
    # Dispatch to format specific functions
    prepare_merge_hook = f"_prepare_merge_{input_store.__class__.__name__}"
    if prepare_merge_hook in globals():
        globals()[prepare_merge_hook](
            input_store, output_store, template_store, **kwargs
        )

    # Generate an index so we can search by source string and location later on
    input_store.makeindex()
    if template_store:
        template_store.makeindex()


def _store_pre_merge(input_store, output_store, template_store, **kwargs):
    """Initialize the new file with things like headers and metadata."""
    # formats that implement poheader interface are a special case
    if isinstance(input_store, poheader.poheader):
        _do_poheaders(input_store, output_store, template_store)
    elif isinstance(input_store, catkeys.CatkeysFile):
        # FIXME: also this should be a format specific hook
        if template_store is not None:
            output_store.header = template_store.header
        else:
            output_store.header = input_store.header

    # Dispatch to format specific functions
    store_pre_merge_hook = f"_store_pre_merge_{input_store.__class__.__name__}"
    if store_pre_merge_hook in globals():
        globals()[store_pre_merge_hook](
            input_store, output_store, template_store, **kwargs
        )


def _store_post_merge(input_store, output_store, template_store, **kwargs):
    """
    Close file after merging all translations, used for adding statistics,
    obsolete messages and similar wrapup tasks.
    """
    # Dispatch to format specific functions
    store_post_merge_hook = f"_store_post_merge_{input_store.__class__.__name__}"
    if store_post_merge_hook in globals():
        globals()[store_post_merge_hook](
            input_store, output_store, template_store, **kwargs
        )


def _unit_post_merge(input_unit, input_store, output_store, template_store, **kwargs):
    """
    Handle any unit level cleanup and situations not handled by the merge()
    function.
    """
    # dispatch to format specific functions
    unit_post_merge_hook = f"_unit_post_merge_{input_unit.__class__.__name__}"
    if unit_post_merge_hook in globals():
        globals()[unit_post_merge_hook](
            input_unit, input_store, output_store, template_store, **kwargs
        )


## Format specific functions
def _unit_post_merge_pounit(input_unit, input_store, output_store, template_store):
    """PO format specific plural string initializtion logic."""
    # FIXME: do we want to do that for poxliff also?
    if input_unit.hasplural() and len(input_unit.target) == 0:
        # untranslated plural unit; Let's ensure that we have the correct
        # number of plural forms:
        nplurals, _plural = output_store.getheaderplural()
        if nplurals and nplurals.isdigit() and nplurals != "2":
            input_unit.target = multistring([""] * int(nplurals))


def _store_post_merge_pofile(input_store, output_store, template_store):
    """PO format specific: adds newly obsoleted messages to end of store."""
    # Let's take care of obsoleted messages
    if template_store:
        newlyobsoleted = []
        for unit in template_store.units:
            if unit.isheader() or unit.isblank():
                continue
            if unit.target and not (
                input_store.findid(unit.getid()) or hasattr(unit, "reused")
            ):
                # Not in .pot, make it obsolete
                unit.makeobsolete()
                newlyobsoleted.append(unit)
        for unit in newlyobsoleted:
            output_store.addunit(unit)


def _do_poheaders(input_store, output_store, template_store):
    """Adds initialized PO headers to output store."""
    # header values
    charset = "UTF-8"
    encoding = "8bit"
    project_id_version = None
    pot_creation_date = None
    po_revision_date = None
    last_translator = None
    language_team = None
    mime_version = None
    plural_forms = None
    kwargs = {}

    if template_store is not None and isinstance(template_store, poheader.poheader):
        templateheadervalues = template_store.parseheader()
        for key, value in templateheadervalues.items():
            if key == "Project-Id-Version":
                project_id_version = value
            elif key == "Last-Translator":
                last_translator = value
            elif key == "Language-Team":
                language_team = value
            elif key == "PO-Revision-Date":
                po_revision_date = value
            elif key in {"POT-Creation-Date", "MIME-Version"}:
                # don't know how to handle these keys, or ignoring them
                pass
            elif key == "Content-Type":
                kwargs[key] = value
            elif key == "Content-Transfer-Encoding":
                encoding = value
            elif key == "Plural-Forms":
                plural_forms = value
            else:
                kwargs[key] = value

    inputheadervalues = input_store.parseheader()
    for key, value in inputheadervalues.items():
        if key in {
            "Project-Id-Version",
            "Last-Translator",
            "Language-Team",
            "PO-Revision-Date",
            "Content-Type",
            "Content-Transfer-Encoding",
            "Plural-Forms",
        }:
            # want to carry these from the template so we ignore them
            pass
        elif key == "POT-Creation-Date":
            pot_creation_date = value
        elif key == "MIME-Version":
            mime_version = value
        else:
            kwargs[key] = value

    output_header = output_store.init_headers(
        charset=charset,
        encoding=encoding,
        project_id_version=project_id_version,
        pot_creation_date=pot_creation_date,
        po_revision_date=po_revision_date,
        last_translator=last_translator,
        language_team=language_team,
        mime_version=mime_version,
        plural_forms=plural_forms,
        **kwargs,
    )

    # Get the header comments and fuzziness state
    # override some values from input file
    if template_store is not None:
        template_header = template_store.header()
        if template_header is not None:
            if template_header.getnotes("translator"):
                output_header.addnote(
                    template_header.getnotes("translator"),
                    "translator",
                    position="replace",
                )
            output_header.markfuzzy(template_header.isfuzzy())


def main(argv=None):
    from translate.convert import convert

    formats = {
        "pot": ("po", convertpot),
        ("pot", "po"): ("po", convertpot),
        "xlf": ("xlf", convertpot),
        ("xlf", "xlf"): ("xlf", convertpot),
        "xliff": ("xliff", convertpot),
        ("xliff", "xliff"): ("xliff", convertpot),
        "ts": ("ts", convertpot),
        "lang": ("lang", convertpot),
        ("lang", "lang"): ("lang", convertpot),
        ("ts", "ts"): ("ts", convertpot),
        "catkeys": ("catkeys", convertpot),
        ("catkeys", "catkeys"): ("catkeys", convertpot),
    }
    parser = convert.ConvertOptionParser(
        formats,
        usepots=True,
        usetemplates=True,
        allowmissingtemplate=True,
        description=__doc__,
    )

    parser.add_option(
        "",
        "--tm",
        dest="tm",
        default=None,
        help="The file to use as translation memory when fuzzy matching",
    )
    parser.passthrough.append("tm")

    defaultsimilarity = 75
    parser.add_option(
        "-s",
        "--similarity",
        dest="min_similarity",
        default=defaultsimilarity,
        type="float",
        help="The minimum similarity for inclusion (default: %d%%)" % defaultsimilarity,
    )
    parser.passthrough.append("min_similarity")

    parser.add_option(
        "--nofuzzymatching",
        dest="fuzzymatching",
        action="store_false",
        default=True,
        help="Disable fuzzy matching",
    )
    parser.passthrough.append("fuzzymatching")

    parser.run(argv)


if __name__ == "__main__":
    main()



#
# Copyright 2010 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
This module represents the Punjabi language.

.. seealso:: :wp:`Punjabi_language`
"""

import re

from translate.lang import common


class pa(common.Common):
    """This class represents Punjabi."""

    sentenceend = "।!?…"

    sentencere = re.compile(
        rf"""(?s)    # make . also match newlines
                            .*?         # anything, but match non-greedy
                            [{sentenceend}]        # the puntuation for sentence ending
                            \s+         # the spacing after the puntuation
                            (?=[^a-z\d])# lookahead that next part starts with
                                        # caps
                            """,
        re.VERBOSE,
    )

    puncdict = {
        ". ": "। ",
        ".\n": "।\n",
    }

    ignoretests = {
        "all": ["simplecaps", "startcaps"],
    }



#
# Copyright 2008-2009 Zuza Software Foundation
#
# This file is part of translate.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Wrapper to launch the bundled CherryPy server."""

import logging

from cheroot.wsgi import Server

logger = logging.getLogger(__name__)


def launch_server(host, port, app, **kwargs):
    """Use cheroot WSGI server, a multithreaded scallable server."""
    server = Server((host, port), app, **kwargs)
    logger.info("Starting server, listening on port %s", port)
    try:
        server.start()
    except KeyboardInterrupt:
        server.stop()



"""tests decoration handling functions that are used by checks."""

from translate.filters import decoration


def test_spacestart():
    """Test operation of spacestart()."""
    assert decoration.spacestart("  Start") == "  "
    assert decoration.spacestart("\u0020\u00a0Start") == "\u0020\u00a0"
    # non-breaking space
    assert decoration.spacestart("\u00a0\u202fStart") == "\u00a0\u202f"
    # Some exotic spaces
    assert (
        decoration.spacestart(
            "\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200aStart"
        )
        == "\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200a"
    )


def test_isvalidaccelerator():
    """Test the isvalidaccelerator() function."""
    # Mostly this tests the old code path where acceptlist is None
    assert not decoration.isvalidaccelerator("")
    assert decoration.isvalidaccelerator("a")
    assert decoration.isvalidaccelerator("1")
    assert not decoration.isvalidaccelerator("ḽ")
    # Test new code path where we actually have an acceptlist
    assert decoration.isvalidaccelerator("a", "aeiou")
    assert decoration.isvalidaccelerator("ḽ", "ḓṱḽṋṅ")
    assert not decoration.isvalidaccelerator("a", "ḓṱḽṋṅ")


def test_find_marked_variables():
    """
    Check that we can identify variables correctly, the first returned
    value is the start location, the second returned value is the actual
    variable sans decoations.
    """
    variables = decoration.findmarkedvariables("The <variable> string", "<", ">")
    assert variables == [(4, "variable")]
    variables = decoration.findmarkedvariables("The $variable string", "$", 1)
    assert variables == [(4, "v")]
    variables = decoration.findmarkedvariables("The $variable string", "$", None)
    assert variables == [(4, "variable")]
    variables = decoration.findmarkedvariables("The $variable string", "$", 0)
    assert variables == [(4, "")]
    variables = decoration.findmarkedvariables("The &variable; string", "&", ";")
    assert variables == [(4, "variable")]
    variables = decoration.findmarkedvariables(
        "The &variable.variable; string", "&", ";"
    )
    assert variables == [(4, "variable.variable")]


def test_getnumbers():
    """Test operation of getnumbers()."""
    assert decoration.getnumbers("") == []
    assert decoration.getnumbers("No numbers") == []
    assert decoration.getnumbers("Nine 9 nine") == ["9"]
    assert decoration.getnumbers("Two numbers: 2 and 3") == ["2", "3"]
    assert decoration.getnumbers("R5.99") == ["5.99"]
    # TODO fix these so that we are able to consider locale specific numbers
    # assert decoration.getnumbers("R5,99") == ["5.99"]
    # assert decoration.getnumbers("1\u00a0000,99") == ["1000.99"]
    assert decoration.getnumbers("36°") == ["36°"]
    assert decoration.getnumbers("English 123, Bengali \u09e7\u09e8\u09e9") == [
        "123",
        "\u09e7\u09e8\u09e9",
    ]


def test_getfunctions():
    """Test operation of getfunctions()."""
    assert decoration.getfunctions("") == []
    assert decoration.getfunctions("There is no function") == []
    assert decoration.getfunctions("Use the getfunction() function.") == [
        "getfunction()"
    ]
    assert decoration.getfunctions(
        "Use the getfunction1() function or the getfunction2() function."
    ) == ["getfunction1()", "getfunction2()"]
    assert decoration.getfunctions("The module.getfunction() method") == [
        "module.getfunction()"
    ]
    assert decoration.getfunctions("The module->getfunction() method") == [
        "module->getfunction()"
    ]
    assert decoration.getfunctions("The module::getfunction() method") == [
        "module::getfunction()"
    ]
    assert decoration.getfunctions("The function().function() function") == [
        "function().function()"
    ]
    assert decoration.getfunctions("Deprecated, use function().") == ["function()"]
    assert decoration.getfunctions("Deprecated, use function() or other().") == [
        "function()",
        "other()",
    ]



from io import BytesIO

from translate.convert import po2tmx
from translate.misc.xml_helpers import XML_NS
from translate.storage import tmx

from . import test_convert


class TestPO2TMX:
    @staticmethod
    def po2tmx(posource, sourcelanguage="en", targetlanguage="af", comment=None):
        """Helper that converts po source to tmx source without requiring files."""
        inputfile = BytesIO(posource.encode("utf-8"))
        outputfile = BytesIO()
        outputfile.tmxfile = tmx.tmxfile(inputfile=None, sourcelanguage=sourcelanguage)
        po2tmx.convertpo(
            inputfile,
            outputfile,
            templatefile=None,
            sourcelanguage=sourcelanguage,
            targetlanguage=targetlanguage,
            comment=comment,
        )
        return outputfile.tmxfile

    def test_basic(self):
        minipo = r"""# Afrikaans translation of program ABC
#
msgid ""
msgstr ""
"Project-Id-Version: program 2.1-branch\n"
"Report-Msgid-Bugs-To: \n"
"POT-Creation-Date: 2006-01-09 07:15+0100\n"
"PO-Revision-Date: 2004-03-30 17:02+0200\n"
"Last-Translator: Zuza Software Foundation <xxx@translate.org.za>\n"
"Language-Team: Afrikaans <translate-discuss-xxx@lists.sourceforge.net>\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"

# Please remember to do something
#: ../dir/file.xml.in.h:1 ../dir/file2.xml.in.h:4
msgid "Applications"
msgstr "Toepassings"
"""
        tmx = self.po2tmx(minipo)
        print("The generated xml:")
        print(bytes(tmx))
        assert tmx.translate("Applications") == "Toepassings"
        assert tmx.translate("bla") is None
        xmltext = bytes(tmx).decode("utf-8")
        assert xmltext.index('creationtool="Translate Toolkit"')
        assert xmltext.index("adminlang")
        assert xmltext.index("creationtoolversion")
        assert xmltext.index("datatype")
        assert xmltext.index("o-tmf")
        assert xmltext.index("segtype")
        assert xmltext.index("srclang")

    def test_sourcelanguage(self):
        minipo = 'msgid "String"\nmsgstr "String"\n'
        tmx = self.po2tmx(minipo, sourcelanguage="xh")
        print("The generated xml:")
        print(bytes(tmx))
        header = tmx.document.find("header")
        assert header.get("srclang") == "xh"

    def test_targetlanguage(self):
        minipo = 'msgid "String"\nmsgstr "String"\n'
        tmx = self.po2tmx(minipo, targetlanguage="xh")
        print("The generated xml:")
        print(bytes(tmx))
        tuv = tmx.document.findall(".//{}".format(tmx.namespaced("tuv")))[1]
        # tag[0] will be the source, we want the target tuv
        assert tuv.get(f"{{{XML_NS}}}lang") == "xh"

    def test_multiline(self):
        """Test multiline po entry."""
        minipo = r'''msgid "First part "
"and extra"
msgstr "Eerste deel "
"en ekstra"'''
        tmx = self.po2tmx(minipo)
        print("The generated xml:")
        print(bytes(tmx))
        assert tmx.translate("First part and extra") == "Eerste deel en ekstra"

    def test_escapednewlines(self):
        """Test the escaping of newlines."""
        minipo = r"""msgid "First line\nSecond line"
msgstr "Eerste lyn\nTweede lyn"
"""
        tmx = self.po2tmx(minipo)
        print("The generated xml:")
        print(bytes(tmx))
        assert tmx.translate("First line\nSecond line") == "Eerste lyn\nTweede lyn"

    def test_escapedtabs(self):
        """Test the escaping of tabs."""
        minipo = r"""msgid "First column\tSecond column"
msgstr "Eerste kolom\tTweede kolom"
"""
        tmx = self.po2tmx(minipo)
        print("The generated xml:")
        print(bytes(tmx))
        assert (
            tmx.translate("First column\tSecond column") == "Eerste kolom\tTweede kolom"
        )

    def test_escapedquotes(self):
        """Test the escaping of quotes (and slash)."""
        minipo = r"""msgid "Hello \"Everyone\""
msgstr "Good day \"All\""

msgid "Use \\\"."
msgstr "Gebruik \\\"."
"""
        tmx = self.po2tmx(minipo)
        print("The generated xml:")
        print(bytes(tmx))
        assert tmx.translate('Hello "Everyone"') == 'Good day "All"'
        assert tmx.translate(r"Use \".") == r"Gebruik \"."

    def test_exclusions(self):
        """Test that empty and fuzzy messages are excluded."""
        minipo = r"""#, fuzzy
msgid "One"
msgstr "Een"

msgid "Two"
msgstr ""

msgid ""
msgstr "Drie"
"""
        tmx = self.po2tmx(minipo)
        print("The generated xml:")
        print(bytes(tmx))
        assert b"<tu" not in bytes(tmx)
        assert len(tmx.units) == 0

    def test_nonascii(self):
        """Tests that non-ascii conversion works."""
        minipo = """msgid "Bézier curve"
msgstr "Bézier-kurwe"
"""
        tmx = self.po2tmx(minipo)
        print(bytes(tmx))
        assert tmx.translate("Bézier curve") == "Bézier-kurwe"

    def test_nonecomments(self):
        """Tests that none comments are imported."""
        minipo = """#My comment rules
msgid "Bézier curve"
msgstr "Bézier-kurwe"
"""
        tmx = self.po2tmx(minipo)
        print(bytes(tmx))
        unit = tmx.findunits("Bézier curve")
        assert len(unit[0].getnotes()) == 0

    def test_otherscomments(self):
        """Tests that others comments are imported."""
        minipo = """#My comment rules
msgid "Bézier curve"
msgstr "Bézier-kurwe"
"""
        tmx = self.po2tmx(minipo, comment="others")
        print(bytes(tmx))
        unit = tmx.findunits("Bézier curve")
        assert unit[0].getnotes() == "My comment rules"

    def test_sourcecomments(self):
        """Tests that source comments are imported."""
        minipo = """#: ../PuzzleFourSided.h:45
msgid "Bézier curve"
msgstr "Bézier-kurwe"
"""
        tmx = self.po2tmx(minipo, comment="source")
        print(bytes(tmx))
        unit = tmx.findunits("Bézier curve")
        assert unit[0].getnotes() == "../PuzzleFourSided.h:45"

    def test_typecomments(self):
        """Tests that others comments are imported."""
        minipo = """#, csharp-format
msgid "Bézier curve"
msgstr "Bézier-kurwe"
"""
        tmx = self.po2tmx(minipo, comment="type")
        print(bytes(tmx))
        unit = tmx.findunits("Bézier curve")
        assert unit[0].getnotes() == "csharp-format"


class TestPO2TMXCommand(test_convert.TestConvertCommand, TestPO2TMX):
    """Tests running actual po2tmx commands on files."""

    convertmodule = po2tmx

    expected_options = [
        "-l LANG, --language=LANG",
        "--source-language=LANG",
        "--comments",
    ]



from io import BytesIO

import pytest
from pytest import mark

from translate.storage import po, xliff
from translate.tools import pomerge


def test_str2bool():
    """Test the str2bool function."""
    assert pomerge.str2bool("yes")
    assert pomerge.str2bool("true")
    assert pomerge.str2bool("1")
    assert not pomerge.str2bool("no")
    assert not pomerge.str2bool("false")
    assert not pomerge.str2bool("0")
    with pytest.raises(ValueError):
        pomerge.str2bool("2")


class TestPOMerge:
    xliffskeleton = """<?xml version="1.0" ?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
  <file original="filename.po" source-language="en-US" datatype="po">
    <body>
        %s
    </body>
  </file>
</xliff>"""

    @staticmethod
    def mergestore(
        templatesource,
        inputsource,
        mergeblanks="yes",
        mergefuzzy="yes",
        mergecomments="yes",
    ):
        """Merges the sources of the given files and returns a new pofile object."""
        templatefile = BytesIO(templatesource.encode())
        inputfile = BytesIO(inputsource.encode())
        outputfile = BytesIO()
        assert pomerge.mergestore(
            inputfile,
            outputfile,
            templatefile,
            mergeblanks=mergeblanks,
            mergefuzzy=mergefuzzy,
            mergecomments=mergecomments,
        )
        outputpostring = outputfile.getvalue()
        return po.pofile(outputpostring)

    @staticmethod
    def mergexliff(
        templatesource,
        inputsource,
        mergeblanks="yes",
        mergefuzzy="yes",
        mergecomments="yes",
    ):
        """Merges the sources of the given files and returns a new xlifffile object."""
        templatefile = BytesIO(templatesource.encode())
        inputfile = BytesIO(inputsource.encode())
        outputfile = BytesIO()
        assert pomerge.mergestore(
            inputfile,
            outputfile,
            templatefile,
            mergeblanks=mergeblanks,
            mergefuzzy=mergefuzzy,
            mergecomments=mergecomments,
        )
        outputxliffstring = outputfile.getvalue()
        print("Generated XML:")
        print(outputxliffstring)
        return xliff.xlifffile(outputxliffstring)

    @staticmethod
    def countunits(pofile):
        """Returns the number of non-header items."""
        if pofile.units[0].isheader():
            return len(pofile.units) - 1
        return len(pofile.units)

    def singleunit(self, pofile):
        """Checks that the pofile contains a single non-header unit, and returns it."""
        assert self.countunits(pofile) == 1
        return pofile.units[-1]

    def test_mergesore_bad_data(self):
        """Test that we catch bad options sent to mergestore."""
        templatefile = BytesIO(b"")
        inputfile = BytesIO(b"")
        outputfile = BytesIO()
        with pytest.raises(ValueError):
            pomerge.mergestore(inputfile, outputfile, templatefile, mergeblanks="yay")
        with pytest.raises(ValueError):
            pomerge.mergestore(inputfile, outputfile, templatefile, mergecomments="yay")

    def test_simplemerge(self):
        """Checks that a simple po entry merges OK."""
        templatepo = """#: simple.test\nmsgid "Simple String"\nmsgstr ""\n"""
        inputpo = """#: simple.test\nmsgid "Simple String"\nmsgstr "Dimpled Ring"\n"""
        pofile = self.mergestore(templatepo, inputpo)
        pounit = self.singleunit(pofile)
        assert pounit.source == "Simple String"
        assert pounit.target == "Dimpled Ring"

    def test_simplemerge_no_locations(self):
        """Checks that a simple po entry merges OK."""
        templatepo = '''msgid "Simple String"
msgstr ""'''
        inputpo = '''msgid "Simple String"
msgstr "Dimpled Ring"'''
        pofile = self.mergestore(templatepo, inputpo)
        pounit = self.singleunit(pofile)
        assert pounit.source == "Simple String"
        assert pounit.target == "Dimpled Ring"

    def test_replacemerge(self):
        """Checks that a simple po entry merges OK."""
        templatepo = (
            """#: simple.test\nmsgid "Simple String"\nmsgstr "Dimpled Ring"\n"""
        )
        inputpo = """#: simple.test\nmsgid "Simple String"\nmsgstr "Dimpled King"\n"""
        pofile = self.mergestore(templatepo, inputpo)
        pounit = self.singleunit(pofile)
        assert pounit.source == "Simple String"
        assert pounit.target == "Dimpled King"

    def test_merging_blanks(self):
        """By default we will merge blanks, but we can also override that."""
        templatepo = (
            """#: simple.test\nmsgid "Simple String"\nmsgstr "Dimpled Ring"\n"""
        )
        inputpo = """#: simple.test\nmsgid "Simple String"\nmsgstr ""\n"""
        pofile = self.mergestore(templatepo, inputpo)
        pounit = self.singleunit(pofile)
        assert pounit.source == "Simple String"
        assert pounit.target == ""
        pofile = self.mergestore(templatepo, inputpo, mergeblanks="no")
        pounit = self.singleunit(pofile)
        assert pounit.source == "Simple String"
        assert pounit.target == "Dimpled Ring"

    def test_merging_fuzzies(self):
        """By default we will merge fuzzies, but can can also override that."""
        templatepo = (
            """#: simple.test\nmsgid "Simple String"\nmsgstr "Dimpled Ring"\n"""
        )
        inputpo = """#: simple.test\n#, fuzzy\nmsgid "Simple String"\nmsgstr "changed fish"\n"""
        pofile = self.mergestore(templatepo, inputpo)
        pounit = self.singleunit(pofile)
        assert pounit.source == "Simple String"
        assert pounit.target == "changed fish"
        assert pounit.isfuzzy()
        pofile = self.mergestore(templatepo, inputpo, mergefuzzy="no")
        pounit = self.singleunit(pofile)
        assert pounit.source == "Simple String"
        assert pounit.target == "Dimpled Ring"
        assert not pounit.isfuzzy()

    def test_merging_locations(self):
        """
        Check that locations on separate lines are output in Gettext form
        of all on one line.
        """
        templatepo = (
            """#: location.c:1\n#: location.c:2\nmsgid "Simple String"\nmsgstr ""\n"""
        )
        inputpo = """#: location.c:1\n#: location.c:2\nmsgid "Simple String"\nmsgstr "Dimpled Ring"\n"""
        expectedpo = f"""#: location.c:1{po.lsep}location.c:2\nmsgid "Simple String"\nmsgstr "Dimpled Ring"\n"""
        pofile = self.mergestore(templatepo, inputpo)
        print(pofile)
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_unit_missing_in_template_with_locations(self):
        """If the unit is missing in the template we should raise an error."""
        templatepo = '''#: location.c:1
msgid "Simple String"
msgstr ""'''
        inputpo = '''#: location.c:1
msgid "Simple String"
msgstr "Dimpled Ring"

#: location.c:1
msgid "Extra string"
msgstr "Perplexa ring"'''
        expectedpo = """#: location.c:1
msgid "Simple String"
msgstr "Dimpled Ring"
"""
        pofile = self.mergestore(templatepo, inputpo)
        print(pofile)
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_unit_missing_in_template_no_locations(self):
        """If the unit is missing in the template we should raise an error."""
        templatepo = '''msgid "Simple String"
msgstr ""'''
        inputpo = '''msgid "Simple String"
msgstr "Dimpled Ring"

msgid "Extra string"
msgstr "Perplexa ring"'''
        expectedpo = """msgid "Simple String"
msgstr "Dimpled Ring"
"""
        pofile = self.mergestore(templatepo, inputpo)
        print(pofile)
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_reflowed_source_comments(self):
        """
        Ensure that we don't duplicate source comments (locations) if they
        have been reflowed.
        """
        templatepo = (
            """#: newMenu.label\n#: newMenu.accesskey\nmsgid "&New"\nmsgstr ""\n"""
        )
        newpo = """#: newMenu.label newMenu.accesskey\nmsgid "&New"\nmsgstr "&Nuwe"\n"""
        expectedpo = f"""#: newMenu.label{po.lsep}newMenu.accesskey\nmsgid "&New"\nmsgstr "&Nuwe"\n"""
        pofile = self.mergestore(templatepo, newpo)
        self.singleunit(pofile)
        print(pofile)
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_comments_with_blank_lines(self):
        """Ensure that we don't loose empty newlines in comments."""
        templatepo = """# # ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# bla bla
msgid "bla"
msgstr "blabla"
"""
        newpo = templatepo
        expectedpo = templatepo
        pofile = self.mergestore(templatepo, newpo)
        self.singleunit(pofile)
        print(pofile)
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_merge_dont_delete_unassociated_comments(self):
        """
        Ensure that we do not delete comments in the PO file that are not
        assocaited with a message block.
        """
        templatepo = """# Lonely comment\n\n# Translation comment\nmsgid "Bob"\nmsgstr "Toolmaker"\n"""
        mergepo = """# Translation comment\nmsgid "Bob"\nmsgstr "Builder"\n"""
        expectedpo = """# Lonely comment\n# Translation comment\nmsgid "Bob"\nmsgstr "Builder"\n"""
        pofile = self.mergestore(templatepo, mergepo)
        #        pounit = self.singleunit(pofile)
        print(pofile)
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_preserve_format_trailing_newlines(self):
        """Test that we can merge messages correctly that end with a newline."""
        templatepo = """msgid "Simple string\\n"\nmsgstr ""\n"""
        mergepo = """msgid "Simple string\\n"\nmsgstr "Dimpled ring\\n"\n"""
        expectedpo = """msgid "Simple string\\n"\nmsgstr "Dimpled ring\\n"\n"""
        pofile = self.mergestore(templatepo, mergepo)
        print(f"Expected:\n{expectedpo}\n\nMerged:\n{bytes(pofile)}")
        assert bytes(pofile).decode("utf-8") == expectedpo

        templatepo = """msgid ""\n"Simple string\\n"\nmsgstr ""\n"""
        mergepo = """msgid ""\n"Simple string\\n"\nmsgstr ""\n"Dimpled ring\\n"\n"""
        expectedpo = """msgid ""\n"Simple string\\n"\nmsgstr "Dimpled ring\\n"\n"""
        expectedpo2 = """msgid "Simple string\\n"\nmsgstr "Dimpled ring\\n"\n"""
        pofile = self.mergestore(templatepo, mergepo)
        print(f"Expected:\n{expectedpo}\n\nMerged:\n{bytes(pofile)}")
        assert (
            bytes(pofile).decode("utf-8") == expectedpo or bytes(pofile) == expectedpo2
        )

    def test_preserve_format_minor_start_and_end_of_sentence_changes(self):
        """
        Test that we are not too fussy about large diffs for simple
        changes at the start or end of a sentence.
        """
        templatepo = """msgid "Target type:"\nmsgstr "Doelsoort"\n\n"""
        mergepo = """msgid "Target type:"\nmsgstr "Doelsoort:"\n"""
        expectedpo = mergepo
        pofile = self.mergestore(templatepo, mergepo)
        print(f"Expected:\n{expectedpo}\n\nMerged:\n{bytes(pofile)}")
        assert bytes(pofile).decode("utf-8") == expectedpo

        templatepo = """msgid "&Select"\nmsgstr "Kies"\n\n"""
        mergepo = """msgid "&Select"\nmsgstr "&Kies"\n"""
        expectedpo = mergepo
        pofile = self.mergestore(templatepo, mergepo)
        print(f"Expected:\n{expectedpo}\n\nMerged:\n{bytes(pofile)}")
        assert bytes(pofile).decode("utf-8") == expectedpo

        templatepo = """msgid "en-us, en"\nmsgstr "en-us, en"\n"""
        mergepo = """msgid "en-us, en"\nmsgstr "af-za, af, en-za, en-gb, en-us, en"\n"""
        expectedpo = mergepo
        pofile = self.mergestore(templatepo, mergepo)
        print(f"Expected:\n{expectedpo}\n\nMerged:\n{bytes(pofile)}")
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_preserve_format_last_entry_in_a_file(self):
        """
        The last entry in a PO file is usualy not followed by an empty
        line.  Test that we preserve this.
        """
        templatepo = """msgid "First"\nmsgstr ""\n\nmsgid "Second"\nmsgstr ""\n"""
        mergepo = (
            """msgid "First"\nmsgstr "Eerste"\n\nmsgid "Second"\nmsgstr "Tweede"\n"""
        )
        expectedpo = (
            """msgid "First"\nmsgstr "Eerste"\n\nmsgid "Second"\nmsgstr "Tweede"\n"""
        )
        pofile = self.mergestore(templatepo, mergepo)
        print(f"Expected:\n{expectedpo}\n\nMerged:\n{bytes(pofile)}")
        assert bytes(pofile).decode("utf-8") == expectedpo

        templatepo = """msgid "First"\nmsgstr ""\n\nmsgid "Second"\nmsgstr ""\n\n"""
        mergepo = (
            """msgid "First"\nmsgstr "Eerste"\n\nmsgid "Second"\nmsgstr "Tweede"\n"""
        )
        expectedpo = (
            """msgid "First"\nmsgstr "Eerste"\n\nmsgid "Second"\nmsgstr "Tweede"\n"""
        )
        pofile = self.mergestore(templatepo, mergepo)
        print(f"Expected:\n{expectedpo}\n\nMerged:\n{bytes(pofile)}")
        assert bytes(pofile).decode("utf-8") == expectedpo

    @mark.xfail(reason="Not Implemented")
    def test_escape_tabs(self):
        """Ensure that input tabs are escaped in the output, like gettext does."""
        # The strings below contains the tab character, not spaces.
        templatepo = """msgid "First	Second"\nmsgstr ""\n\n"""
        mergepo = """msgid "First	Second"\nmsgstr "Eerste	Tweede"\n"""
        expectedpo = r"""msgid "First\tSecond"
msgstr "Eerste\tTweede"
"""
        pofile = self.mergestore(templatepo, mergepo)
        print(f"Expected:\n{expectedpo}\n\nMerged:\n{bytes(pofile)}")
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_preserve_comments_layout(self):
        """
        Ensure that when we merge with new '# (poconflict)' or other
        comments that we don't mess formating.
        """
        templatepo = """#: filename\nmsgid "Desktop Background.bmp"\nmsgstr "Desktop Background.bmp"\n\n"""
        mergepo = """# (pofilter) unchanged: please translate\n#: filename\nmsgid "Desktop Background.bmp"\nmsgstr "Desktop Background.bmp"\n"""
        expectedpo = mergepo
        pofile = self.mergestore(templatepo, mergepo)
        print(f"Expected:\n{expectedpo}\n\nMerged:\n{bytes(pofile)}")
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_merge_dos2unix(self):
        """Test that merging a comment line with dos newlines doesn't add a new line."""
        templatepo = """# User comment\n# (pofilter) Translate Toolkit comment\n#. Automatic comment\n#: location_comment.c:110\nmsgid "File"\nmsgstr "File"\n\n"""
        mergepo = """# User comment\r\n# (pofilter) Translate Toolkit comment\r\n#. Automatic comment\r\n#: location_comment.c:110\r\nmsgid "File"\r\nmsgstr "Ifayile"\r\n\r\n"""
        expectedpo = """# User comment\n# (pofilter) Translate Toolkit comment\n#. Automatic comment\n#: location_comment.c:110\nmsgid "File"\nmsgstr "Ifayile"\n"""
        pofile = self.mergestore(templatepo, mergepo)
        assert bytes(pofile).decode("utf-8") == expectedpo

        # Unassociated comment
        templatepo = """# Lonely comment\n\n#: location_comment.c:110\nmsgid "Bob"\nmsgstr "Toolmaker"\n"""
        mergepo = """# Lonely comment\r\n\r\n#: location_comment.c:110\r\nmsgid "Bob"\r\nmsgstr "Builder"\r\n\r\n"""
        expectedpo = """# Lonely comment\n#: location_comment.c:110\nmsgid "Bob"\nmsgstr "Builder"\n"""
        pofile = self.mergestore(templatepo, mergepo)
        assert bytes(pofile).decode("utf-8") == expectedpo

        # New comment
        templatepo = """#: location_comment.c:110\nmsgid "File"\nmsgstr "File"\n\n"""
        mergepo = """# User comment\r\n# (pofilter) Translate Toolkit comment\r\n#: location_comment.c:110\r\nmsgid "File"\r\nmsgstr "Ifayile"\r\n\r\n"""
        expectedpo = """# User comment\n# (pofilter) Translate Toolkit comment\n#: location_comment.c:110\nmsgid "File"\nmsgstr "Ifayile"\n"""
        pofile = self.mergestore(templatepo, mergepo)
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_xliff_into_xliff(self):
        templatexliff = (
            self.xliffskeleton
            % """<trans-unit>
        <source>red</source>
        <target></target>
</trans-unit>"""
        )
        mergexliff = (
            self.xliffskeleton
            % """<trans-unit>
        <source>red</source>
        <target>rooi</target>
</trans-unit>"""
        )
        xlifffile = self.mergexliff(templatexliff, mergexliff)
        assert len(xlifffile.units) == 1
        unit = xlifffile.units[0]
        assert unit.source == "red"
        assert unit.target == "rooi"

    def test_po_into_xliff(self):
        templatexliff = (
            self.xliffskeleton
            % """<trans-unit>
        <source>red</source>
        <target></target>
</trans-unit>"""
        )
        mergepo = 'msgid "red"\nmsgstr "rooi"'
        xlifffile = self.mergexliff(templatexliff, mergepo)
        assert len(xlifffile.units) == 1
        unit = xlifffile.units[0]
        assert unit.source == "red"
        assert unit.target == "rooi"

    def test_xliff_into_po(self):
        templatepo = '# my comment\nmsgid "red"\nmsgstr ""'
        mergexliff = (
            self.xliffskeleton
            % """<trans-unit>
        <source>red</source>
        <target>rooi</target>
</trans-unit>"""
        )
        expectedpo = '# my comment\nmsgid "red"\nmsgstr "rooi"\n'
        pofile = self.mergestore(templatepo, mergexliff)
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_merging_dont_merge_kde_comments_found_in_translation(self):
        """If we find a KDE comment in the translation (target) then do not merge it."""
        templatepo = """msgid "_: KDE comment\\n"\n"File"\nmsgstr "File"\n\n"""
        mergepo = """msgid "_: KDE comment\\n"\n"File"\nmsgstr "_: KDE comment\\n"\n"Ifayile"\n\n"""
        expectedpo = """msgid ""\n"_: KDE comment\\n"\n"File"\nmsgstr "Ifayile"\n"""
        pofile = self.mergestore(templatepo, mergepo)
        print(f"Expected:\n{expectedpo}\n\nMerged:\n{bytes(pofile)}")
        assert bytes(pofile).decode("utf-8") == expectedpo

        # Translated kde comment.
        mergepo = """msgid "_: KDE comment\\n"\n"File"\nmsgstr "_: KDE kommentaar\\n"\n"Ifayile"\n\n"""
        print(f"Expected:\n{expectedpo}\n\nMerged:\n{bytes(pofile)}")
        assert bytes(pofile).decode("utf-8") == expectedpo

        # multiline KDE comment
        templatepo = """msgid "_: KDE "\n"comment\\n"\n"File"\nmsgstr "File"\n\n"""
        mergepo = """msgid "_: KDE "\n"comment\\n"\n"File"\nmsgstr "_: KDE "\n"comment\\n"\n"Ifayile"\n\n"""
        expectedpo = """msgid ""\n"_: KDE comment\\n"\n"File"\nmsgstr "Ifayile"\n"""
        pofile = self.mergestore(templatepo, mergepo)
        print(f"Expected:\n{expectedpo}\n\nMerged:\n{bytes(pofile)}")
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_merging_untranslated_with_kde_disambiguation(self):
        """Test merging untranslated messages that are the same except for KDE disambiguation."""
        templatepo = r"""#: sendMsgTitle
#: sendMsgTitle.accesskey
msgid "_: sendMsgTitle sendMsgTitle.accesskey\n"
"Send Message"
msgstr ""

#: sendMessageCheckWindowTitle
#: sendMessageCheckWindowTitle.accesskey
msgid "_: sendMessageCheckWindowTitle sendMessageCheckWindowTitle.accesskey\n"
"Send Message"
msgstr ""
"""
        mergepo = rf"""#: sendMsgTitle{po.lsep}sendMsgTitle.accesskey
msgid ""
"_: sendMsgTitle sendMsgTitle.accesskey\n"
"Send Message"
msgstr "Stuur"

#: sendMessageCheckWindowTitle{po.lsep}sendMessageCheckWindowTitle.accesskey
msgid ""
"_: sendMessageCheckWindowTitle sendMessageCheckWindowTitle.accesskey\n"
"Send Message"
msgstr "Stuur"
"""
        expectedpo = mergepo
        pofile = self.mergestore(templatepo, mergepo)
        print(f"Expected:\n{expectedpo}\n---\nMerged:\n{bytes(pofile)}\n---")
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_merging_header_entries(self):
        """Check that we do the right thing if we have header entries in the input PO."""
        templatepo = r"""#, fuzzy
msgid ""
msgstr ""
"Project-Id-Version: PACKAGE VERSION\n"
"Report-Msgid-Bugs-To: new@example.com\n"
"POT-Creation-Date: 2006-11-11 11:11+0000\n"
"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\n"
"Last-Translator: FULL NAME <EMAIL@ADDRESS>\n"
"Language-Team: LANGUAGE <LL@li.org>\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"Plural-Forms: nplurals=INTEGER; plural=EXPRESSION;\n"
"X-Generator: Translate Toolkit 0.10rc2\n"

#: simple.test
msgid "Simple String"
msgstr ""
"""
        mergepo = r"""msgid ""
msgstr ""
"Project-Id-Version: Pootle 0.10\n"
"Report-Msgid-Bugs-To: old@example.com\n"
"POT-Creation-Date: 2006-01-01 01:01+0100\n"
"PO-Revision-Date: 2006-09-09 09:09+0900\n"
"Last-Translator: Joe Translate <joe@example.com>\n"
"Language-Team: Pig Latin <piglatin@example.com>\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"Plural-Forms: nplurals=2; plural=(n != 1);\n"
"X-Generator: Translate Toolkit 0.9\n"

#: simple.test
msgid "Simple String"
msgstr "Dimpled Ring"
"""
        expectedpo = r"""msgid ""
msgstr ""
"Project-Id-Version: Pootle 0.10\n"
"Report-Msgid-Bugs-To: new@example.com\n"
"POT-Creation-Date: 2006-11-11 11:11+0000\n"
"PO-Revision-Date: 2006-09-09 09:09+0900\n"
"Last-Translator: Joe Translate <joe@example.com>\n"
"Language-Team: Pig Latin <piglatin@example.com>\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"Plural-Forms: nplurals=2; plural=(n != 1);\n"
"X-Generator: Translate Toolkit 0.10rc2\n"

#: simple.test
msgid "Simple String"
msgstr "Dimpled Ring"
"""
        pofile = self.mergestore(templatepo, mergepo)
        print(f"Expected:\n{expectedpo}\n---\nMerged:\n{bytes(pofile)}\n---")
        assert bytes(pofile).decode("utf-8") == expectedpo

    def test_merging_different_locations(self):
        """
        Test when merging units that are unchanged except for changed
        locations that we don't go fuzzy (bug 1583).
        """
        templatepo = r"""#, fuzzy
msgid ""
msgstr ""
"Content-Type: text/plain; charset=utf-8\n"
"Content-Transfer-Encoding: 8bit\n"

#: sentinelheadline
msgctxt "sentinelheadline"
msgid "DESTROY SENTINELS"
msgstr "ZERSTÖRE WACHPOSTEN"

#: sentinelheadline1
msgctxt "sentinelheadline1"
msgid "DESTROY SENTINELS"
msgstr "ZERSTÖRE WACHPOSTEN"
"""
        mergepo = r"""msgid ""
msgstr ""
"Content-Type: text/plain; charset=utf-8\n"
"Content-Transfer-Encoding: 8bit\n"

#: sentinelheadline
#: sentinelheadline1
msgctxt "sentinelheadline"
msgid "DESTROY SENTINELS"
msgstr "ZERSTÖRE WACHPOSTEN"

#: sentinelheadline1
msgctxt "sentinelheadline1"
msgid "DESTROY SENTINELS"
msgstr "ZERSTÖRE WACHPOSTEN"
"""
        expectedpo2 = r"""msgid ""
msgstr ""
"Content-Type: text/plain; charset=utf-8\n"
"Content-Transfer-Encoding: 8bit\n"

#: sentinelheadline sentinelheadline1
msgctxt "sentinelheadline"
msgid "DESTROY SENTINELS"
msgstr "ZERSTÖRE WACHPOSTEN"

#: sentinelheadline1
msgctxt "sentinelheadline1"
msgid "DESTROY SENTINELS"
msgstr "ZERSTÖRE WACHPOSTEN"
"""

        expectedpo = mergepo
        pofile = self.mergestore(templatepo, mergepo)
        output = bytes(pofile).decode("utf-8")
        print(f"Expected:\n{expectedpo}\n---\nMerged:\n{output}\n---")
        assert output in {expectedpo, expectedpo2}
